"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""

# import cv2.cv as cv
import os
from lmfit import Model, Parameters
from lmfit.models import VoigtModel, GaussianModel
from sklearn.metrics import r2_score, mean_squared_error
import pickle
import fabio
# import pyFAI
from os.path import isfile, exists
from os import makedirs
from ..utils.file_manager import fullPath, getBlankImageAndMask
from ..utils.histogram_processor import *
from ..utils.image_processor import *
import musclex

class EquatorImage:
    """
    A class for Bio-Muscle processing - go to process() to see all processing steps
    """
    def __init__(self, dir_path, filename):
        """
        Initial value for EquatorImage object
        :param dir_path: directory path of input image
        :param filename: image file name
        """
        self.sigmaS = 0.0001
        self.dir_path = dir_path
        self.filename = filename
        self.orig_img = fabio.open(fullPath(dir_path, filename)).data
        self.rotated_img = None
        self.version = musclex.__version__
        cache = self.loadCache()
        if cache is None:
            # info dictionary will save all results
            self.info = {
                "mask_thres" : getMaskThreshold(self.orig_img)
            }
        else:
            self.info = cache

    def process(self, settings):
        """
        All processing steps - all settings are provided by bio-muscle app as a dictionary
        settings must have ...
        nPeaks - number of peaks (int)
        model - "Voigt" or "Gaussian" (str)
        sigmac - (float)
        isSkeletal - is it skeletal muscle (boolean)
        """
        self.updateInfo(settings)
        self.applyBlankAndMask()
        self.findCenter()
        self.getRotationAngle()
        self.calculateRmin()
        self.getIntegrateArea()
        self.getHistogram()
        self.applyConvexhull()
        self.getPeaks()
        self.managePeaks()
        self.fitModel()
        if "no_cache" not in settings:
            self.saveCache()

    def removeInfo(self, k=None):
        """
        Remove information from info dictionary by k as a key. If k is None, remove all information in the dictionary
        :param k: key of dictionary
        :return: -
        """

        if k is None:
            keys = list(self.info.keys())
            for k in keys:
                del self.info[k]
        else:
            if k in self.info.keys(): # remove from dictionary if the key exists
                del self.info[k]

    def updateInfo(self, settings):
        if settings['orientation_model'] is None:
            if 'orientation_model' not in self.info or self.info['orientation_model'] is None:
                settings['orientation_model'] = 0
            else:
                del settings['orientation_model']
        self.info.update(settings)

    def applyBlankAndMask(self):
        """
        Subtract the original image with blank image and set pixels in mask below the mask threshold
        """
        img = np.array(self.orig_img, dtype='float32')
        if self.info['blank_mask']:
            blank, mask = getBlankImageAndMask(self.dir_path)
            if blank is not None:
                img = img - blank
            if mask is not None:
                img[mask>0] = self.info['mask_thres']-1

        self.image = img

    def findCenter(self):
        """
       Find center of the diffraction. The center will be kept in self.info["center"].
       Once the center is calculated, the rotation angle will be re-calculated, so self.info["rotationAngle"] is deleted
       """
        print("Center is being calculated...")
        if 'center' not in self.info:
            self.info['center'] = getCenter(self.orig_img)
            self.removeInfo('rotationAngle') # Remove rotationAngle from info dict to make it be re-calculated
        print("Done. Center is" + str(self.info['center']))

    def getRotationAngle(self):
        """
        Find rotation angle of the diffraction. Turn the diffraction equator to be horizontal. The angle will be kept in self.info["rotationAngle"]
        Once the rotation angle is calculated, the rmin will be re-calculated, so self.info["rmin"] is deleted
        """
        print("Rotation Angle is being calculated...")
        if 'rotationAngle' not in self.info:
            if "fixed_angle" in self.info:
                self.info['rotationAngle'] = self.info["fixed_angle"]
            else:
                center = self.info['center']
                img = copy.copy(self.image)
                self.info['rotationAngle'] = getRotationAngle(img, center, self.info['orientation_model'])
            self.removeInfo('rmin')  # Remove R-min from info dict to make it be re-calculated

        print("Done. Rotation Angle is " + str(self.info['rotationAngle']))

    def calculateRmin(self):
        """
        Calculate R-min of the diffraction. The R-min will be kept in self.info["rmin"]
        Once the R-min is calculated, the integrated area (Box Width) will be re-calculated, so self.info["int_area"] is deleted
        """
        if 'fixed_rmin' in self.info:
            self.info['rmin'] = self.info['fixed_rmin']
            print("R-min is fixed as " + str(self.info['rmin']))
            return

        print("R-min is being calculated...")
        if 'rmin' not in self.info:
            img = copy.copy(self.orig_img)
            center = self.info['center']

            if img.shape == (1043, 981):
                det = "pilatus1m"
            else:
                det = "agilent_titan"

            corners = [(0, 0), (img.shape[1], 0), (0, img.shape[0]), (img.shape[1], img.shape[0])]
            npt_rad = int(round(max([distance(center, c) for c in corners])))
            ai = pyFAI.AzimuthalIntegrator(detector=det)
            ai.setFit2D(100, center[0], center[1])
            tth, I = ai.integrate1d(img, npt_rad, unit="r_mm") # Get 1D Azimuthal integrated histogram
            self.info['rmin'] = getFirstVallay(I) # R-min is value before the first valley
            self.removeInfo('int_area')  # Remove integrated area from info dict to make it be re-calculated
            #
            # fig = plt.figure()
            # ax = fig.add_subplot(111)
            # ax.plot(I[:int(len(I)/4)])
            # ax.axvline(self.info['rmin'], color = 'r')
            # ax.set_xlabel("radius (pixel)")
            # ax.set_ylabel("intensity (pixel)")
            # ax.set_xlim((0, len(I[:int(len(I)/4)])))
            # fig.show()

        print("Done. R-min is " + str(self.info['rmin']))

    def getRotatedImage(self, img=None, angle=None):
        """
        Get rotated image by angle. If the input params are not specified. image = original input image, angle = self.info["rotationAngle"]
        :param img: input image
        :param angle: rotation angle
        :return: rotated image
        """

        if img is None:
            img = copy.copy(self.image)
        if angle is None:
            angle = self.info['rotationAngle']
        if '90rotation' in self.info and self.info['90rotation'] is True:
            angle = angle - 90 if angle > 90 else angle + 90

        if self.rotated_img is None or self.rotated_img[0] != self.info["center"] or self.rotated_img[1] != self.info["rotationAngle"] or (self.rotated_img[2] != img).any():
            # encapsulate rotated image for using later as a list of [center, angle, original image, rotated image[
            self.rotated_img = [self.info["center"], angle, img,
                                rotateImage(img, self.info["center"], angle, self.info['mask_thres'])]

        return self.rotated_img[3]

    def getIntegrateArea(self):
        """
        Calculate Integrated Area (Box width) of the diffraction. The integrated area will be start and end points in y direction for getting histogram.
        The Integrated Area will be kept in self.info["int_area"]
        Once the Integrated Area is calculated, the histograms will be re-calculated, so self.info["hist"] is deleted
        """
        print("Integrated Area is being calculated...")
        if 'int_area' not in self.info:
            center = self.info['center']
            if 'fixed_int_area' in self.info: # integrated area is fixed by users
                self.info['int_area'] = self.info['fixed_int_area']
            else:
                rmin = self.info['rmin']
                img = getCenterRemovedImage(copy.copy(self.image), tuple(center), rmin) # remove center location
                rotate_img = self.getRotatedImage(img) # rotate image
                init_range = int(round(rmin * 1.5)) # specify initial guess by using 150% or R-min
                top = max(0, center[1] - init_range)
                bottom = min(center[1] + init_range, rotate_img.shape[0])
                area = rotate_img[top:bottom, :]
                hist = np.sum(area, axis=1)  # Get the horizontal histogram from the initital area
                hull = convexHull(hist) # Apply convexull

                if len(hist) > 0:
                    max_loc = np.argmax(hull)
                    r = 1
                    l = 1
                    # Find 0 on the left and right
                    while max_loc - l >= 0 and max_loc + r < len(hull) and (
                                    hull[max_loc - l] != 0 or hull[max_loc + r] != 0):
                        if hull[max_loc - l] != 0:
                            l += 1
                        if hull[max_loc + r] != 0:
                            r += 1

                    if max_loc + r < center[1] or max_loc - l > center[1] or abs(r + l) < rmin * .7:
                        self.info['int_area'] = (
                        int(round(center[1] - rmin * .7)), int(round(center[1] + rmin * .7)) + 1)
                    else:
                        center = center[1] - rmin + max_loc
                        self.info['int_area'] = (int(round(center - l * 1.2)), int(round(center + r * 1.2)) + 1)
                else:
                    # if convex hull does not work properly, integration area will be 70% of R-min
                    self.info['int_area'] = (int(round(center[1] - rmin * .7)), int(round(center[1] + rmin * .7)) + 1)

            self.removeInfo('hist') # Remove histograms from info dict to make it be re-calculated

        print("Done. Integrated Area is " + str(self.info['int_area']))

    def getHistogram(self):
        """
        Getting original histogram of the diffraction in the integrated area. Histogram will be kept in self.info["hist"]
        Once getting histogram is done, the background subtracted histogram will be re-calculated, so self.info["hulls"] is deleted
        """
        print("Getting Histogram...")
        if 'hist' not in self.info:
            int_area = self.info['int_area']
            img = self.getRotatedImage()
            self.info['hist'] = np.sum(img[int_area[0]:int_area[1], :], axis=0)
            self.removeInfo('hulls')  # Remove background subtracted histogram from info dict to make it be re-calculated

        print("Done.")

    def applyConvexhull(self):
        """
       Getting backgound subtracted histogram by applying Convex hull algorithm. Background Subtracted Histogram will be kept in self.info["hulls"]
       This will provide left, right, and both separated by centerX
       Once getting background subtracted histogram is done, the temp peaks will be re-calculated, so self.info["tmp_peaks"] is deleted
       """
        print("Applying Convexhull...")
        if 'hulls' not in self.info:
            center = self.info['center']
            shapes = self.image.shape
            rmax = int(min(center[0], center[1], shapes[1] - center[0], shapes[0] - center[1]) * 0.8)
            rmin = self.info['rmin']
            hist = copy.copy(self.info['hist'])
            int_area = self.info['int_area']
            img_area = self.getRotatedImage()[int_area[0]:int_area[1], :]
            ignore = np.array([any(img_area[:, i] <= self.info['mask_thres']) for i in range(img_area.shape[1])])
            if any(ignore):
                left_ignore = ignore[:center[0]]
                left_ignore = left_ignore[::-1]
                right_ignore = ignore[center[0]:]
            else:
                left_ignore = None
                right_ignore = None

            left_hist = hist[:center[0]][::-1]
            right_hist = hist[center[0]:]
            right_hull = convexHull(right_hist, start_p=rmin, end_p=rmax, ignore=right_ignore) # Apply Convex hull for right histogram
            left_hull = convexHull(left_hist, start_p=rmin, end_p=rmax, ignore=left_ignore) # Apply Convex hull for left histogram

            hull = copy.copy(list(left_hull[::-1]))
            hull.extend(list(right_hull[:]))

            self.info['hulls'] = {'right': right_hull,
                                  'left': left_hull,
                                  'all': hull}

            self.info['hists'] = {'right': right_hist,
                                  'left': left_hist,
                                  'all': hist}

            self.removeInfo('tmp_peaks') # Remove temp peaks from info dict to make it be re-calculated

        print("Done")

    def getPeaks(self):
        """
        Finding Temp Peaks from background subtracted histogram. Temp peaks are just peaks which are found by algorithm.
        This might miss some peaks or detect to many peaks because of noise. These peaks will be managed again in the next step.
        Temp peaks will be kept in self.info["tmp_peaks"].
        Once getting Temp peaks is done, the real peaks will be re-calculated, so self.info["peaks"] is deleted
        """
        print("Finding Peaks...")
        if 'tmp_peaks' not in self.info:
            left_peaks = getPeaksFromHist(self.info['hulls']['left'])
            right_peaks = getPeaksFromHist(self.info['hulls']['right'])
            self.info['tmp_peaks'] = {'left': left_peaks, 'right': right_peaks}
            self.removeInfo('peaks')  # Remove real peaks from info dict to make it be re-calculated

        print("Done. Peaks are found : " + str(self.info['tmp_peaks']))


    def managePeaks(self):
        """
        Getting real peaks from temp peaks. Temp peaks will be considered as real peaks if it in the Hexagonal Pattern location.
        Real peaks will be kept in self.info["peaks"].
        Once getting real peaks is done, the fitting results will be re-calculated, so self.info["fit_results"] is deleted
        """
        print("Model Peaks are being selected...")
        if 'peaks' not in self.info:
            left_peaks = movePeaks(self.info['hulls']['left'], sorted(self.info['tmp_peaks']['left']), 5)
            right_peaks = movePeaks(self.info['hulls']['right'], sorted(self.info['tmp_peaks']['right']), 5)
            first_left, first_right = self.findFirstSymmetricPeaks(left_peaks, right_peaks)

            self.removeInfo('fit_results')  # Remove fit results from info dict to make it be re-calculated

            if first_left is None:
                print('WARNING: '+str(self.filename) + '- no effective peaks detected. Model will not be fit.')
                return

            self.hexagonalPattern(first_left, first_right, left_peaks, right_peaks)
        print("Done. Selected Peaks are" + str(self.info['peaks']))

    def findFirstSymmetricPeaks(self, left_peaks, right_peaks):
        """
        Get first symatric peaks from left and right
        :param left_peaks: peaks on the left histogram (list)
        :param right_peaks: peaks on the right histogram (list)
        :return: first symmetric peaks, return None if there are no symmetric peaks
        """
        dist_thres = 20 # Threshold for difference between distance of left and right peaks
        for lp in left_peaks:
            for rp in right_peaks:
                if abs(lp - rp) < dist_thres:
                    return lp, rp
        return None, None

    def hexagonalPattern(self, first_left, first_right, left_peaks, right_peaks):
        """
        Set all peaks information after apply Hexagonal Pattern including ..
        - S values i.e S10, S11, S20, ...
        - Peaks
        - Missed Peaks

        :param first_left: first symmetric peak on the left (int)
        :param first_right: first symmetric peak on the right (int)
        :param left_peaks: peaks on the left histogram (list)
        :param right_peaks: peaks on the right histogram  (list)
        :return:
        """
        if "nPeaks" not in self.info:
            self.info["nPeaks"] = 2
        allpeaks = {'left': [], 'right': []}
        missed_peaks = {'left': [], 'right': []}
        allpeaks['left'].append(first_left)
        allpeaks['right'].append(first_right)
        max_loc = min(len(self.info["hulls"]['right']), len(self.info["hulls"]['left']))

        # self.info['peak'] = [symP, peak]
        S10 = int((first_left + first_right) / 2.)

        self.info['S'] = [S10]

        # Find other S values based on S10 and theta function
        SList = []
        for i in range(1, int(self.info["nPeaks"])):
            s_pos = int(np.round(S10 * theta(i)))
            if s_pos > max_loc:
                break
            SList.append(s_pos)

        maximum_nPeaks = 2

        # Find peaks correponding to S values
        for i, S in enumerate(SList):
            self.info['S'].append(S)
            if i + 2 > self.info["nPeaks"]:
                break
            SLeft = min(left_peaks, key=lambda p: abs(p - S))
            SRight = min(right_peaks, key=lambda p: abs(p - S))
            if abs(SLeft - S) > 10 or abs(SRight - S) > 10 or SLeft in allpeaks['left'] or SRight in allpeaks['right']:
                missed_peaks['left'].append(S)
                missed_peaks['right'].append(S)
                allpeaks['left'].append(S)
                allpeaks['right'].append(S)
            else:
                if SLeft not in allpeaks['left'] and SRight not in allpeaks['right']:
                    allpeaks['left'].append(SLeft)
                    allpeaks['right'].append(SRight)
            maximum_nPeaks += 2

        allpeaks['left'] = allpeaks['left'][0:int(maximum_nPeaks / 2)]
        allpeaks['right'] = allpeaks['right'][0:int(maximum_nPeaks / 2)]
        self.info['S'] = self.info['S'][0:int(maximum_nPeaks / 2)]
        # self.info['S20'] = center[0]+int(S20)

        self.info['peaks'] = allpeaks
        # self.info['nPeaks'] = maximum_nPeaks
        self.info['MissedPeaks'] = missed_peaks

    def fitModel(self):
        """
        Fit model to background subtracted histogram by using S10, peak location as initial guess
        Fit results will be kept in self.info["fit_results"].
        """
        if 'peaks' not in self.info:
            # model cannot be fitted if peaks are not found
            return
        print("Fitting Model ...")
        if 'fit_results' not in self.info:
            left_hull = self.info['hulls']['left']
            right_hull = self.info['hulls']['right']
            left_peaks = self.info['peaks']['left']
            right_peaks = self.info['peaks']['right']

            S = self.info['S']

            left_widths = self.getPeakWidths('left')
            right_widths = self.getPeakWidths('right')

            right_height = [right_hull[right_peaks[i]] for i in range(len(S))]
            left_height = [left_hull[left_peaks[i]] for i in range(len(S))]

            # init sigma D
            sigmaD = max(left_widths)

            # initial sigma S
            sigmaS = self.sigmaS

            # init areas
            left_areas = [left_height[i] * left_widths[i] * np.sqrt(2 * np.pi) for i in range(len(left_peaks))]
            right_areas = [right_height[i] * right_widths[i] * np.sqrt(2 * np.pi) for i in range(len(right_peaks))]

            hull_hist = self.info['hulls']['all']
            x = np.arange(0, len(hull_hist))
            histNdarray = np.array(hull_hist)
            total_area = sum(histNdarray)
            centerX = self.info['center'][0]

            margin = 10.

            # Add all fitting variables to params
            params = Parameters()
            params.add("centerX", centerX, min=centerX - margin, max=centerX + margin)
            params.add("S10", S[0], min=S[0] - margin, max=S[0] + margin)

            for i in range(len(left_areas)):
                params.add("left_area" + str(i + 1), max(left_areas[i], 100), min=0)
            for i in range(len(right_areas)):
                params.add("right_area" + str(i + 1), max(right_areas[i],100), min=0)

            # Add independent variables to int_vars
            int_vars = {
                'x': x,
                'left_sigmac': self.info['left_sigmac'],
                'right_sigmac': self.info['right_sigmac'],
                'model': self.info["model"],
                'isSkeletal': self.info['isSkeletal'],
                # 'gamma': 1.0
            }

            # Set initial parameters or independent parameters on each side
            for side in ['left', 'right']:

                # If params are fixed, add them to int_vars
                if side+'_fix_sigmas' in self.info.keys():
                    int_vars[side+'_sigmas'] = self.info[side+'_fix_sigmas']
                else:
                    params.add(side+'_sigmas', sigmaS,  min=-10., max=10)

                if side+'_fix_sigmad' in self.info.keys():
                    int_vars[side+'_sigmad'] = self.info[side+'_fix_sigmad']
                else:
                    params.add(side+'_sigmad', sigmaD, min=0, max=30.)

                if side+'_fix_gamma' in self.info.keys():
                    int_vars[side+'_gamma'] = self.info[side+'_fix_gamma']
                else:
                    init_gamma = np.sqrt(self.info['left_sigmac'] ** 2 + (sigmaD * theta(1)) ** 2 + (sigmaS * (theta(1) ** 2)) ** 2)
                    params.add(side+'_gamma', init_gamma, min=0, max=init_gamma * 5.0 + 1.)

                if self.info['isSkeletal']:
                    if side+'_fix_zline' in self.info:
                        int_vars[side+'_zline'] = self.info[side+'_fix_zline']
                    else:
                        init_z = 1.5
                        params.add(side+'_zline', S[0] * init_z, min=S[0] * init_z - 10, max=S[0] * init_z + 10)

                    if side+'_fix_sigz' in self.info:
                        int_vars[side+'_sigmaz'] = self.info[side+'_fix_sigz']
                    else:
                        params.add(side+'_sigmaz', 8., min=0., max=20.)

                    if side+'_fix_intz' in self.info:
                        int_vars[side+'_intz'] = self.info[side+'_fix_intz']
                    else:
                        init_intz = max(left_areas[0] / 5., right_areas[0] / 5.)
                        params.add(side+'_intz', init_intz, min=0, max=init_intz*4.+1.)

                    if side+'_fix_gammaz' in self.info:
                        int_vars[side+'_gammaz'] = self.info[side+'_fix_gammaz']
                    else:
                        params.add(side+'_gammaz', 8., min=-5., max=30.)
                else:
                    # set all z line variables as independent
                    int_vars[side+'_zline'] = 0
                    int_vars[side+'_sigmaz'] = 0
                    int_vars[side+'_intz'] = 0
                    int_vars[side+'_gammaz'] = 0

            # Bias K
            if 'fix_k' in self.info:
                int_vars['k'] = self.info['fix_k']
            else:
                params.add('k', 0., min=-1, max=max(histNdarray.max(),1.))

            # Fit model
            model = Model(cardiacFit, nan_policy='propagate', independent_vars=int_vars.keys())
            min_err = 999999999
            final_result = None

            # for method in ['leastsq', 'lbfgsb', 'powell', 'cg', 'slsqp', 'nelder', 'cobyla', 'tnc']:
            for method in ['leastsq']:
                result = model.fit(histNdarray, verbose = False, method=method, params=params, **int_vars)
                if result is not None:
                    res = result.values
                    res.update(int_vars)
                    err = mean_squared_error(histNdarray, cardiacFit(**res))
                    if err < min_err:
                        min_err = err
                        final_result = result

            if final_result is not None :
                fit_result = final_result.values
                fit_result.update(int_vars)
                fit_result["fiterror"] = 1. - r2_score(cardiacFit(**fit_result), histNdarray)
                del fit_result['x']
                left_areas = [fit_result['left_area' + str(i + 1)] for i in range(len(left_peaks))]
                right_areas = [fit_result['right_area' + str(i + 1)] for i in range(len(right_peaks))]

                if len(left_areas) < 2 or len(right_areas) < 2:
                    return
                #### Get Ratio between I10 and I11 ####
                fit_result['left_ratio'] = 1.*(left_areas[1] / left_areas[0])
                fit_result['right_ratio'] = 1.*(right_areas[1] / right_areas[0])
                avg_area_ratios = (fit_result['left_ratio'] + fit_result['right_ratio']) / 2.
                fit_result['avg_ratio'] = avg_area_ratios
                fit_result['left_areas'] = left_areas
                fit_result['right_areas'] = right_areas

                centerX = fit_result["centerX"]
                S10 = fit_result["S10"]
                model_peaks = [centerX - S10 * theta(i) for i in range(len(left_peaks))]
                model_peaks.extend([centerX + S10 * theta(i) for i in range(len(right_peaks))])

                fit_result['model_peaks'] = sorted(model_peaks)
                all_S = [S10 * theta(i) for i in range(len(left_peaks))]
                fit_result['all_S'] = all_S

                if "lambda_sdd" in self.info.keys():
                    fit_result['d10'] = self.info["lambda_sdd"] / fit_result['S10']

                self.info['fit_results'] = fit_result
                # print "cardiacFit result : ",result.values

                # original_hist = self.info['rhist']
                # plt.plot(x, original_hist, label = 'histogram_BG')

        if 'fit_results' in self.info:
            print("Done. Fitting Results : " + str(self.info['fit_results']))
            if self.info['fit_results']['fiterror'] > 0.2:
                print("WARNING : High Fitting Error")
        else:
            print("Model cannot be fitted.")

    def getPeakWidths(self, side):
        """
        Get initial peaks' widths from histogram and peaks on speicific side
        :param side: "left" or "right"
        :return: peaks' widths list
        """
        widthList = []
        peaks = self.info['peaks'][side]
        missed = self.info['MissedPeaks'][side]
        hist = self.info['hulls'][side]
        for p in peaks:
            if p in missed:
                # if peak is missing, init width as 7
                widthList.append(7)
            else:
                i, j = self.findPeakRange(p, hist)
                r = max((j - i), 5)
                w = 1. * r / 2.3548
                widthList.append(w)
        return widthList

    def findPeakRange(self, peak, hist):
        """
        Find location of Full width at half maximum of peak
        :param peak: (int)
        :param hist: (list or numpy.array)
        :return: left and right point of Full width at half maximum
        """
        i = j = peak
        midPeakHeight = hist[peak] / 2.
        thresh = np.mean(hist)
        while midPeakHeight <= hist[i] and hist[i] > thresh:
            i -= 1
        while midPeakHeight <= hist[j] and hist[j] > thresh:
            j += 1
        return i, j

    def loadCache(self):
        """
        Load info dict from cache. Cache file will be filename.info in folder "eq_cache"
        :return: cached info (dict)
        """
        cache_path = fullPath(self.dir_path, "eq_cache")
        cache_file = fullPath(cache_path, self.filename + '.info')

        if exists(cache_path) and isfile(cache_file):
            cinfo = pickle.load(open(cache_file, "rb"))
            if cinfo != None:
                if cinfo['program_version'] == self.version:
                    return cinfo
        return None

    def saveCache(self):
        """
        Save info dict to cache. Cache file will be save as filename.info in folder "eq_cache"
        :return: -
        """
        cache_path = fullPath(self.dir_path, "eq_cache")
        cache_file = fullPath(cache_path, self.filename + '.info')

        # Create cache path if it does not exist
        if not exists(cache_path):
            makedirs(cache_path)

        self.info['program_version'] = self.version
        pickle.dump(self.info, open(cache_file, "wb"))

    def delCache(self):
        """
        Delete cache
        :return: -
        """
        cache_path = fullPath(self.dir_path, "eq_cache")
        cache_file = fullPath(cache_path, self.filename + '.info')
        if exists(cache_path) and isfile(cache_file):
            os.remove(cache_file)


def cardiacFit(x, centerX, S10, model, isSkeletal, k
               , left_sigmad, left_sigmas, left_sigmac, left_gamma, left_intz, left_sigmaz, left_zline, left_gammaz
               , right_sigmad, right_sigmas, right_sigmac, right_gamma, right_intz, right_sigmaz, right_zline, right_gammaz, **kwargs):

    """
    Using for fitting model by lmfit
    :param x: x range (list)
    :param centerX: x value of center (int)
    :param S10: (float)
    :param sigmad: (float) for each side
    :param sigmas: (float) for each side
    :param sigmac: (float) for each side
    :param model: "Voigt" or "Gaussian"
    :param gamma: use for Voigt model (float) for each side
    :param isSkeletal: (boolean)
    :param k: linear background
    :param intz: intensity of z line (float) for each side
    :param sigmaz: sigma of z line (float) for each side
    :param zline: center of z line (float) for each side
    :param kwargs: area1, area2, ..., areaN as parameters or areas as a list
    :return:
    """
    # print "center=", centerX
    # print "S10=",S10
    # print "model=",model
    # print "isSkeletal=",isSkeletal
    # print "left_sigmad=", left_sigmad
    # print "left_sigmas=", left_sigmas
    # print "left_sigmac=", left_sigmac
    # print "left_gamma=", left_gamma
    # print "left_intz=", left_intz
    # print "left_sigmaz=", left_sigmaz
    # print "left_zline=", left_zline
    # print "left_gammaz=", left_gammaz
    # print "right_sigmad=", right_sigmad
    # print "right_sigmas=", right_sigmas
    # print "right_sigmac=", right_sigmac
    # print "right_gamma=", right_gamma
    # print "right_intz=", right_intz
    # print "right_sigmaz=", right_sigmaz
    # print "right_zline=", right_zline
    # print "right_gammaz=", right_gammaz
    # print "kwargs=", kwargs
    # print "======================"
    if kwargs is not None:

        if 'left_areas' in kwargs.keys() and 'right_areas' in kwargs.keys() :
            left_areas = kwargs['left_areas']
            right_areas = kwargs['right_areas']
        else:
            left_areas_dict = {}
            right_areas_dict = {}
            for kv in kwargs.items():
                key = kv[0]
                value = kv[1]
                if 'left_area' in key:
                    left_areas_dict[int(key[9:])] = value
                if 'right_area' in key:
                    right_areas_dict[int(key[10:])] = value

            left_areas = sorted(left_areas_dict.items(), key=lambda kv: kv[0])
            left_areas = [v for (_, v) in left_areas]
            right_areas = sorted(right_areas_dict.items(), key=lambda kv: kv[0])
            right_areas = [v for (_, v) in right_areas]

        result = cardiacSide(model, 'left', x, centerX, S10, left_sigmac, left_sigmad, left_sigmas, left_gamma, left_areas)
        result += cardiacSide(model, 'right', x, centerX, S10, right_sigmac, right_sigmad, right_sigmas, right_gamma,
                             right_areas)
        if isSkeletal:
            if model == "Gaussian":
                mod = GaussianModel()
                result += mod.eval(x=x, amplitude=left_intz, center=centerX - left_zline,
                                   sigma=left_sigmaz)
                result += mod.eval(x=x, amplitude=right_intz, center=centerX + right_zline,
                                   sigma=right_sigmaz)
            elif model == "Voigt":
                mod = VoigtModel()
                result += mod.eval(x=x, amplitude=left_intz, center=centerX + left_zline,
                                   sigma=left_sigmaz, gamma=left_gammaz)
                result += mod.eval(x=x, amplitude=right_intz, center=centerX - right_zline,
                                   sigma=right_sigmaz, gamma=right_gammaz)

        return result + k

    return 0

def cardiacSide(model, side, x, centerX, S10, sigmac, sigmad, sigmas, gamma, areas):
    for i, area in enumerate(areas):
        if side == 'left':
            hk = i
            p = centerX - S10 * theta(hk)
        else:
            hk = i
            p = centerX + S10 * theta(hk)

        sigmahk = np.sqrt(sigmac ** 2 + (sigmad * theta(hk)) ** 2 + (sigmas * (theta(hk) ** 2)) ** 2)

        if model == "Gaussian":
            # if i == 0:
            #     result = (areas[i]/sigmahk)*np.exp(-np.log(2)*((x-p)/sigmahk)**2)
            # else:
            #     result += (areas[i]/sigmahk)*np.exp(-np.log(2)*((x-p)/sigmahk)**2)
            mod = GaussianModel()
            if i == 0:
                result = mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk)
            else:
                result += mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk)

        elif model == "Voigt":
            mod = VoigtModel()
            if i == 0:
                result = mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk, gamma=gamma)
            else:
                result += mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk, gamma=gamma)

    return result

def cardiacFit_old(x, centerX, S10, sigmad, sigmas, sigmac, model, gamma, isSkeletal, intz, sigmaz, zline, gammaz, **kwargs):
    """
    Using for fitting model by lmfit
    :param x: x range (list)
    :param centerX: x value of center (int)
    :param S10: (float)
    :param sigmad: (float)
    :param sigmas: (float)
    :param sigmac: (float)
    :param model: "Voigt" or "Gaussian"
    :param gamma: use for Voigt model (float)
    :param isSkeletal: (boolean)
    :param intz: intensity of z line
    :param sigmaz: sigma of z line
    :param zline: center of z line
    :param kwargs: area1, area2, ..., areaN as parameters or areas as a list
    :return:
    """
    if kwargs is not None:

        result = None

        if 'areas' in kwargs.keys():
            areas = kwargs['areas']
        else:
            areas_dict = {}
            for kv in kwargs.items():
                if 'area' in kv[0]:
                    areas_dict[int(kv[0][4:])] = kv[1]

            areas = sorted(areas_dict.items(), key=lambda kv: kv[0])
            areas = [v for (k, v) in areas]

        nPeaks = int(len(areas)/2)

        for i, area in enumerate(areas):
            if i < nPeaks:
                hk = (nPeaks) - i - 1
                p = centerX - S10 * theta(hk)
            else:
                hk = i - (nPeaks)
                p = centerX + S10 * theta(hk)

            sigmahk = np.sqrt(sigmac ** 2 + (sigmad * theta(hk)) ** 2 + (sigmas * (theta(hk) ** 2)) ** 2)

            if model == "Gaussian":
                # if i == 0:
                #     result = (areas[i]/sigmahk)*np.exp(-np.log(2)*((x-p)/sigmahk)**2)
                # else:
                #     result += (areas[i]/sigmahk)*np.exp(-np.log(2)*((x-p)/sigmahk)**2)
                mod = GaussianModel()
                if i == 0:
                    result = mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk)
                else:
                    result += mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk)

            elif model == "Voigt":
                mod = VoigtModel()
                if i == 0:
                    result = mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk, gamma=gamma)
                else:
                    result += mod.eval(x=x, amplitude=areas[i], center=p, sigma=sigmahk, gamma=gamma)

        if isSkeletal:
            if model == "Gaussian":
                mod = GaussianModel()
                result += mod.eval(x=x, amplitude=intz, center=centerX + zline,
                                   sigma=sigmaz)
                result += mod.eval(x=x, amplitude=intz, center=centerX - zline,
                                   sigma=sigmaz)
            elif model == "Voigt":
                mod = VoigtModel()
                result += mod.eval(x=x, amplitude=intz, center=centerX + zline,
                                   sigma=sigmaz, gamma=gammaz)
                result += mod.eval(x=x, amplitude=intz, center=centerX - zline,
                                   sigma=sigmaz, gamma=gammaz)

        return result

    return 0


def getCardiacGraph(x, fit_results):
    plot_params = {
        'centerX': fit_results['centerX'],
        'S10': fit_results['S10'],
        'model': fit_results['model'],
        'isSkeletal': fit_results['isSkeletal'],
        'left_areas': fit_results['left_areas'],
        'right_areas': fit_results['right_areas'],

        'left_sigmac': fit_results['left_sigmac'],
        'left_sigmad': fit_results['left_sigmad'],
        'left_sigmas': fit_results['left_sigmas'],
        'left_gamma': fit_results['left_gamma'],
        'left_zline': fit_results['left_zline'],
        'left_sigmaz': fit_results['left_sigmaz'],
        'left_intz': fit_results['left_intz'],
        'left_gammaz': fit_results['left_gammaz'],

        'right_sigmac': fit_results['right_sigmac'],
        'right_sigmad': fit_results['right_sigmad'],
        'right_sigmas': fit_results['right_sigmas'],
        'right_gamma': fit_results['right_gamma'],
        'right_zline': fit_results['right_zline'],
        'right_sigmaz': fit_results['right_sigmaz'],
        'right_intz': fit_results['right_intz'],
        'right_gammaz': fit_results['right_gammaz'],
        'k': fit_results['k']
    }
    return cardiacFit(x=x, **plot_params)

def theta(h, k=-1):
    """
    Theta value that used for distance calculation. S10 * theta
    :param h: index 1
    :param k: index 2, if k is not specified, h will be index count by peak from center
    :return: theta value
    """
    if k == -1:
        # add more coords if necessary
        SList = [(1, 0), (1, 1), (2, 0), (2, 1), (3, 0), (2, 2), (3, 1), (4, 0), (3, 2), (4, 1),
                 (5, 0), (3, 3), (4, 2), (5, 1), (6, 0), (4, 3), (5, 2), (6, 1), (4, 4), (5, 3),
                 (7, 0), (6, 2), (7, 1), (5, 4), (6, 3), (8, 0), (7, 2), (8, 1), (5, 5), (6, 4),
                 (7, 3), (8, 2), (6, 5), (7, 4), (8, 3), (6, 6), (7, 5), (8, 4), (7, 6), (8, 5),
                 (7, 7), (8, 6), (8, 7), (8, 8)]

        hk = SList[h]
        return theta(hk[0], hk[1])
    return np.sqrt(float(h ** 2 + k ** 2 + h * k))



"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""
import matplotlib.pyplot as plt
from .pyqt_utils import *
import matplotlib.patches as patches
import numpy as np
from ..modules.ProjectionProcessor import layerlineModel, layerlineModelBackground, layerlineBackground, meridianBackground, meridianGauss
from ..utils.image_processor import getNewZoom

class ProjectionBoxTab(QWidget):
    """
    Fitting Tabs : left or right
    Display fitting graph and providing options
    """
    def __init__(self, parent, name):
        QWidget.__init__(self)
        self.parent = parent
        self.name = name
        self.function = None
        self.syncUI = False
        self.need_update = True
        self.checkableButtons = []
        self.graphMaxBound = None
        self.zoom1 = None
        self.zoom2 = None
        self.zoomRect = None
        self.centerX = None
        self.initUI()
        self.setAllToolTips()
        self.setConnections()

    def getCenterX(self):
        """
        Get center X
        :return: center X
        """
        if self.parent.projProc is None:
            return 0

        if self.centerX is None:
            info = self.parent.projProc.info
            name = self.name

            box = info['boxes'][name]
            start_x = box[0][0]
            start_y = box[1][0]

            if info['types'][name] == 'h':
                self.centerX = self.parent.projProc.orig_img.shape[1] / 2. - 0.5 - start_x
            else:
                self.centerX = self.parent.projProc.orig_img.shape[0] / 2. - 0.5 - start_y

        return self.centerX

    def initUI(self):
        """
        Initial all GUIs including : 4 plots and result table
        """
        self.setContentsMargins(0, 0, 0, 0)
        self.tabLayout = QHBoxLayout(self)
        self.graphFigure1 = plt.figure()
        self.graphAxes1 = self.graphFigure1.add_subplot(111)
        self.graphCanvas1 = FigureCanvas(self.graphFigure1)

        self.graphFigure2 = plt.figure()
        self.graphAxes2 = self.graphFigure2.add_subplot(111)
        self.graphCanvas2 = FigureCanvas(self.graphFigure2)

        self.optionsFrame = QFrame()
        self.optionsFrame.setFixedWidth(350)
        self.optionsLayout = QVBoxLayout(self.optionsFrame)

        self.displayOptionsGroup = QGroupBox("Display Options")
        self.dispOptLayout = QGridLayout(self.displayOptionsGroup)

        self.histChkBx = QCheckBox("Original Projection")
        self.histChkBx.setChecked(True)
        self.hullRangeChkBx = QCheckBox('Hull Range')
        self.hullRangeChkBx.setEnabled(self.parent.bgsubs[self.name] == 1)
        self.hullRangeChkBx.setChecked(self.parent.bgsubs[self.name]==1)
        self.fitmodelChkBx = QCheckBox("Fitting Model")
        self.fitmodelChkBx.setChecked(True)
        self.bgChkBx = QCheckBox("Background")
        self.bgChkBx.setChecked(True)
        self.maxPeaksChkBx = QCheckBox("Max Peaks")
        self.maxPeaksChkBx.setChecked(False)
        self.peaksChkBx = QCheckBox("Model Peaks")
        self.peaksChkBx.setChecked(True)
        self.centerChkBx = QCheckBox("Center")
        self.centerChkBx.setChecked(False)
        self.subHistChkBx = QCheckBox("Subtracted Projection")
        self.subHistChkBx.setChecked(True)
        self.baselineChkBx = QCheckBox("Baselines")
        self.baselineChkBx.setChecked(False)
        self.centroidChkBx = QCheckBox("Centroids")
        self.centroidChkBx.setChecked(False)

        self.zoomInButton = QPushButton("Zoom in")
        self.zoomInButton.setCheckable(True)
        self.zoomOutButton = QPushButton("Zoom out")
        self.fullButton = QPushButton("Full")
        self.checkableButtons.append(self.zoomInButton)
        self.dispOptLayout.addWidget(self.histChkBx, 0, 0, 1, 1)
        self.dispOptLayout.addWidget(self.fitmodelChkBx, 0, 1, 1, 1)
        self.dispOptLayout.addWidget(self.hullRangeChkBx, 2, 0, 1, 1)
        self.dispOptLayout.addWidget(self.bgChkBx, 2, 1, 1, 1)
        self.dispOptLayout.addWidget(self.maxPeaksChkBx, 3, 0, 1, 1)
        self.dispOptLayout.addWidget(self.peaksChkBx, 3, 1, 1, 1)
        self.dispOptLayout.addWidget(self.centerChkBx, 4, 0, 1, 1)
        self.dispOptLayout.addWidget(self.centroidChkBx, 4, 1, 1, 1)
        self.dispOptLayout.addWidget(self.baselineChkBx, 5, 0, 1, 1)
        self.dispOptLayout.addWidget(self.subHistChkBx, 5, 1, 1, 1)
        self.dispOptLayout.addWidget(self.zoomInButton, 6, 0, 1, 1)
        self.dispOptLayout.addWidget(self.zoomOutButton, 6, 1, 1, 1)
        self.dispOptLayout.addWidget(self.fullButton, 7, 0, 1, 2)

        self.settingGroup = QGroupBox("Setting")
        self.settingLayout = QGridLayout(self.settingGroup)
        self.peaksButton = QPushButton("Select Peaks")
        self.peaksButton.setCheckable(True)
        self.checkableButtons.append(self.peaksButton)
        self.hullRangeButton = QPushButton("Set Manual Convex Hull Range")
        self.hullRangeButton.setCheckable(True)
        self.hullRangeButton.setHidden(self.parent.bgsubs[self.name]!=1)
        box = self.parent.allboxes[self.name]
        width = int(np.ceil(abs(box[0][0]-box[0][1])/2.))

        self.startHull = QSpinBox()
        self.startHull.setRange(0, width)
        self.startHull.setKeyboardTracking(False)
        self.startHull.setHidden(self.parent.bgsubs[self.name] != 1)
        self.endHull = QSpinBox()
        self.endHull.setRange(0, width)
        self.endHull.setKeyboardTracking(False)
        self.endHull.setHidden(self.parent.bgsubs[self.name] != 1)
        self.checkableButtons.append(self.hullRangeButton)
        self.settingLayout.addWidget(self.peaksButton, 0, 0, 1, 2)
        self.settingLayout.addWidget(self.hullRangeButton, 1, 0, 1, 2)
        if self.parent.bgsubs[self.name]== 1:
            self.settingLayout.addWidget(QLabel("Start"), 2, 0, 1, 1, Qt.AlignCenter)
            self.settingLayout.addWidget(QLabel("End"), 2, 1, 1, 1, Qt.AlignCenter)
        self.settingLayout.addWidget(self.startHull, 3, 0, 1, 1)
        self.settingLayout.addWidget(self.endHull, 3, 1, 1, 1)

        self.results_text = QLabel()

        self.resultTable1 = QTableWidget()
        self.resultTable1.setColumnCount(4)
        self.resultTable1.setHorizontalHeaderLabels(["Max Peak", "Gauss Center", "Gauss Sigma", "Gauss Area"])
        self.resultTable1.horizontalHeader().setStretchLastSection(True)
        self.resultTable1.setColumnWidth(0, 75)
        self.resultTable1.setColumnWidth(1, 75)
        self.resultTable1.setColumnWidth(2, 75)
        self.resultTable1.setColumnWidth(3, 75)
        self.resultTable1.setFixedHeight(100)

        self.resultTable2 = QTableWidget()
        self.resultTable2.setColumnCount(3)
        self.resultTable2.setHorizontalHeaderLabels(["Baseline", "Centroid", "Width"])
        self.resultTable2.horizontalHeader().setStretchLastSection(True)
        self.resultTable2.setColumnWidth(0, 100)
        self.resultTable2.setColumnWidth(1, 100)
        self.resultTable2.setColumnWidth(2, 100)
        self.resultTable2.setFixedHeight(100)
        self.pnButtons = QHBoxLayout()
        self.prevButton = QPushButton("<<<")
        self.nextButton = QPushButton(">>>")
        self.pnButtons.addWidget(self.prevButton)
        self.pnButtons.addWidget(self.nextButton)

        self.optionsLayout.addWidget(self.displayOptionsGroup)
        self.optionsLayout.addSpacing(10)
        self.optionsLayout.addWidget(self.settingGroup)
        self.optionsLayout.addSpacing(10)
        self.optionsLayout.addWidget(QLabel("<h3>Fitting Results</h3>"))
        self.optionsLayout.addWidget(self.resultTable1)
        self.optionsLayout.addWidget(QLabel("<h3>Other Results</h3>"))
        self.optionsLayout.addWidget(self.resultTable2)
        self.optionsLayout.addStretch()
        self.optionsLayout.addLayout(self.pnButtons)

        self.graphLayout = QVBoxLayout()
        self.graphLayout.addWidget(self.graphCanvas1)
        self.graphLayout.addWidget(self.graphCanvas2)
        self.tabLayout.addLayout(self.graphLayout)
        self.tabLayout.addWidget(self.optionsFrame)

    def setAllToolTips(self):
        """
        Set Tooltips for widgets
        """
        pass

    def setConnections(self):
        """
        Set connection for interactive widgets
        """
        self.histChkBx.stateChanged.connect(self.resetUI)
        self.fitmodelChkBx.stateChanged.connect(self.resetUI)
        self.hullRangeChkBx.stateChanged.connect(self.resetUI)
        self.bgChkBx.stateChanged.connect(self.resetUI)
        self.maxPeaksChkBx.stateChanged.connect(self.resetUI)
        self.peaksChkBx.stateChanged.connect(self.resetUI)
        self.centerChkBx.stateChanged.connect(self.resetUI)
        self.subHistChkBx.stateChanged.connect(self.resetUI)
        self.baselineChkBx.stateChanged.connect(self.resetUI)
        self.centroidChkBx.stateChanged.connect(self.resetUI)

        self.zoomInButton.clicked.connect(self.zoomInclicked)
        self.zoomOutButton.clicked.connect(self.zoomOutclicked)
        self.fullButton.clicked.connect(self.fullClicked)

        self.peaksButton.clicked.connect(self.addPeaks)
        self.hullRangeButton.clicked.connect(self.setManualHullRange)
        self.startHull.valueChanged.connect(self.hullRangeChanged)
        self.endHull.valueChanged.connect(self.hullRangeChanged)

        self.resultTable2.itemChanged.connect(self.handleItemChanged)

        self.prevButton.clicked.connect(self.parent.prevClicked)
        self.nextButton.clicked.connect(self.parent.nextClicked)

        self.graphFigure1.canvas.mpl_connect('button_press_event', self.graphClicked)
        self.graphFigure2.canvas.mpl_connect('button_press_event', self.graphClicked2)
        self.graphFigure1.canvas.mpl_connect('motion_notify_event', self.graphOnMotion1)
        self.graphFigure2.canvas.mpl_connect('motion_notify_event', self.graphOnMotion2)
        self.graphFigure1.canvas.mpl_connect('button_release_event', self.graphReleased)
        self.graphFigure2.canvas.mpl_connect('button_release_event', self.graphReleased)
        self.graphFigure1.canvas.mpl_connect('figure_leave_event', self.graphReleased)
        self.graphFigure2.canvas.mpl_connect('figure_leave_event', self.graphReleased)


    def hullRangeChanged(self):
        """
        Trigger when convex hull range is changed
        """
        if self.parent.projProc is not None and not self.syncUI:
            self.parent.hull_ranges[self.name] = (self.startHull.value(), self.endHull.value())
            self.parent.projProc.removeInfo(self.name, 'hists2')
            self.parent.processImage()

    def handleItemChanged(self, item):
        """
        Trigger when a item in table is changed
        :param item:
        :return:
        """
        if self.parent.projProc is not None and item.column() == 0 and not self.syncUI:
            self.parent.projProc.setBaseline(self.name, item.row(), item.text())
            self.parent.processImage()

    def zoomInclicked(self):
        """
        Trigger when zoom in button is pressed
        """
        if self.zoomInButton.isChecked():
            self.function = ['zoom']
        else:
            self.function = None
            self.resetUI()

    def zoomOutclicked(self):
        """
        Trigger when zoom out clicked
        :return:
        """
        if self.graphMaxBound is not None:
            xlim = self.graphMaxBound[0]
            ylim = self.graphMaxBound[1]

            def increaseLimit(fig, canvas, ax, xlim, ylim):
                old_xlim = ax.get_xlim()
                old_ylim = ax.get_ylim()
                xscale = abs(old_xlim[0]-old_xlim[1])*0.1
                yscale = abs(old_ylim[0]-old_ylim[1])*0.1
                new_x1 = max(old_xlim[0] - xscale, xlim[0])
                new_x2 = min(old_xlim[1] + xscale, xlim[1])
                new_y1 = max(old_ylim[0] - yscale, ylim[0])
                new_y2 = min(old_ylim[1] + yscale, ylim[1])
                zoom = [(new_x1, new_x2), (new_y1,new_y2)]
                ax.set_xlim(zoom[0])
                ax.set_ylim(zoom[1])
                canvas.draw_idle()
                return zoom

            self.zoom1 = increaseLimit(self.graphFigure1, self.graphCanvas1, self.graphAxes1, xlim, ylim)
            self.zoom2 = increaseLimit(self.graphFigure2, self.graphCanvas2, self.graphAxes2, xlim, ylim)

    def fullClicked(self):
        """
        Triggered when Full button is pressed
        """
        self.zoom1 = None
        self.zoom2 = None
        self.resetUI()

    def setZoomIn(self, point1, point2):
        """
        Set Zoom in for plot n. If n of point1 and point2 are not the same, just refresh UI.
        :param point1: (n, (x1, y1))
        :param point2: (n, (x2, y2))
        :return:
        """
        if point1[0] == point2[0]:
            # Set zoom area for a plot
            pt1 = point1[1]
            pt2 = point2[1]
            x1 = min(pt1[0], pt2[0])
            x2 = max(pt1[0], pt2[0])
            y1 = min(pt1[1], pt2[1])
            y2 = max(pt1[1], pt2[1])
            zoom = [(x1, x2), (y1, y2)]
            if point1[0] == 1:
                self.zoom1 = zoom
            else:
                self.zoom2 = zoom
        self.resetUI()

    def graphReleased(self, event):
        """
        Triggered when mouse is released from plot
        :return:
        """
        if self.function is not None and 'move' in self.function[0]:
            self.function = None
            self.parent.pixel_detail.setText("")

    def graphClicked2(self, event):
        """
        Triggered when the second plot is clicked
        """
        x = event.xdata
        y = event.ydata

        if self.parent.projProc is None or x is None or y is None:
            return

        func = self.function

        if func is None:
            self.function = ['move2', (x, y)]
        elif func[0] == 'zoom':
            # select zoom in area for the second plot
            func.append((2, (x, y)))
            if len(func) == 3:
                self.setZoomIn(func[1], func[2])
        else:
            # call event handler for first plot
            self.graphClicked(event)

    def graphClicked(self, event):
        """
        Triggered when the first plot is clicked
        """
        x = event.xdata
        y = event.ydata
        if self.parent.projProc is None or x is None or y is None:
            return

        if self.function is None:
            self.function = ['move1', (x,y)]
            return

        func = self.function
        box = self.parent.allboxes[self.name]
        type = self.parent.boxtypes[self.name]
        if type == 'h':
            center = self.parent.projProc.orig_img.shape[1] / 2 - 0.5 - box[0][0]
        else:
            center = self.parent.projProc.orig_img.shape[0] / 2 - 0.5 - box[1][0]

        distance = int(round(abs(center - x)))

        if func[0] == 'peaks':
            peaks = func[1]
            peaks.append(distance)

            ax = self.graphAxes1
            ax.axvline(center + distance, color='#ff630a', linewidth=2)
            ax.axvline(center - distance, color='#ff630a', linewidth=2)

            ax = self.graphAxes2
            ax.axvline(center + distance, color='#ff630a', linewidth=2)
            ax.axvline(center - distance, color='#ff630a', linewidth=2)

            self.graphCanvas1.draw_idle()
            self.graphCanvas2.draw_idle()

        elif func[0] == 'hull':
            hull_range = func[1]
            hull_range.append(distance)

            ax = self.graphAxes1
            ax.axvline(center + distance, color='k', linewidth=2)
            ax.axvline(center - distance, color='k', linewidth=2)

            ax = self.graphAxes2
            ax.axvline(center + distance, color='k', linewidth=2)
            ax.axvline(center - distance, color='k', linewidth=2)

            self.graphCanvas1.draw_idle()
            self.graphCanvas2.draw_idle()

            if len(hull_range) == 2:
                hull_range = tuple(sorted(hull_range))
                self.parent.hull_ranges[self.name] = hull_range
                self.parent.projProc.removeInfo(self.name, 'hists2')
                self.parent.processImage()
        elif func[0] == 'zoom':
            # select zoom in area for the second plot
            func.append((1, (x, y)))
            if len(func) == 3:
                self.setZoomIn(func[1], func[2])

    def drawRectangle(self, fig, canvas, ax, point1, point2):
        """
        Draw a rectangle
        :param fig:
        :param canvas:
        :param point1:
        :param point2:
        :return:
        """
        if self.zoomRect is not None and self.zoomRect in ax.patches:
            ax.patches.remove(self.zoomRect)

        x1 = min(point1[0], point2[0])
        x2 = max(point1[0], point2[0])
        y1 = min(point1[1], point2[1])
        y2 = max(point1[1], point2[1])
        w = abs(x1-x2)
        h = abs(y1-y2)

        self.zoomRect = patches.Rectangle((x1, y1), w, h,
                                           linewidth=1, edgecolor='r', facecolor='none', linestyle='dotted')
        ax.add_patch(self.zoomRect)
        canvas.draw_idle()

    def graphOnMotion1(self, event):
        """
        Trigger when mouse hovers the first plot
        """
        x = event.xdata
        y = event.ydata
        if x is not None and y is not None:
            centerX = self.getCenterX()
            distance = x - centerX
            hist = self.parent.projProc.info['hists'][self.name]
            self.parent.pixel_detail.setText("Distance = " + str(round(distance, 3))+", Intensity = "+str(hist[int(round(x))]))
            if self.function is not None:
                if self.function[0] == 'move1' and self.graphMaxBound is not None and self.zoom1 is not None:
                    # change zoom-in location to move around plot
                    func = self.function
                    ax = self.graphAxes1
                    move = (func[1][0] - x, func[1][1] - y)
                    self.zoom1 = getNewZoom(self.zoom1, move, self.graphMaxBound[0][1], self.graphMaxBound[1][1], self.graphMaxBound[1][0])
                    ax.set_xlim(self.zoom1[0])
                    ax.set_ylim(self.zoom1[1])
                    self.graphCanvas1.draw_idle()
                elif self.function[0] == 'zoom' and len(self.function) == 2 and self.function[1][0] == 1:
                    # Draw rectangle
                    start_pt = self.function[1][1]
                    self.drawRectangle(self.graphFigure1, self.graphCanvas1, self.graphAxes1, start_pt, (x, y))



    def graphOnMotion2(self, event):
        """
        Trigger when mouse hovers the first plot
        """
        x = event.xdata
        y = event.ydata
        if x is not None and y is not None:
            centerX = self.getCenterX()
            distance = x - centerX
            all_hists =  self.parent.projProc.info['subtracted_hists']
            if self.name in all_hists:
                hist = all_hists[self.name]
                self.parent.pixel_detail.setText(
                    "Distance = " + str(round(distance, 3)) + ", Intensity = " + str(hist[int(round(x))]))
            else:
                self.parent.pixel_detail.setText("Distance = " + str(round(distance, 3)))
            if self.function is not None:
                if self.function[0] == 'move2' and self.graphMaxBound is not None and self.zoom2 is not None:
                    # change zoom-in location to move around plot
                    func = self.function
                    ax = self.graphAxes2
                    move = (func[1][0] - x, func[1][1] - y)
                    self.zoom2 = getNewZoom(self.zoom2, move, self.graphMaxBound[0][1], self.graphMaxBound[1][1], self.graphMaxBound[1][0])
                    ax.set_xlim(self.zoom2[0])
                    ax.set_ylim(self.zoom2[1])
                    self.graphCanvas2.draw_idle()
                elif self.function[0] == 'zoom' and len(self.function) == 2 and self.function[1][0] == 2:
                    # Draw rectangle
                    start_pt = self.function[1][1]
                    self.drawRectangle(self.graphFigure2, self.graphCanvas2, self.graphAxes2, start_pt, (x,y))

    def setManualHullRange(self):
        """
        Trigger when "Set Manual Convex hull ranges" is pressed
        :return:
        """
        if self.hullRangeButton.isChecked():
            self.function = ['hull', []]
        else:
            self.resetUI()

    def addPeaks(self):
        """
        Trigger when "Select Peaks" is pressed
        :return:
        """
        if self.peaksButton.isChecked():
            self.peaksButton.setText("Done")
            self.function = ['peaks', []]
        else:
            self.peaksButton.setText("Select Peaks")
            peaks = self.function[1]
            self.function = None
            self.parent.addPeakstoBox(self.name, peaks)

    def keyPressEvent(self, event):
        """
        Manage key press event on keyboard
        """
        key = event.key()

        if key == Qt.Key_Escape:
            self.resetUI()

        self.parent.keyPressEvent(event)

    def clearFlags(self):
        """
        clear all flags
        :return:
        """
        self.need_update = True
        self.function = None
        self.peaksButton.setText("Select Peaks")
        for b in self.checkableButtons:
            self.syncUI = True
            b.setChecked(False)
            self.syncUI = False

    def resetUI(self):
        """
        clear flags and update UI
        :return:
        """
        self.clearFlags()
        self.updateUI()

    def updateUI(self):
        """
        Draw plots and display results in text
        :return:
        """
        if self.parent.projProc is None or not self.need_update:
            return

        self.syncUI = True

        # Update graphs
        info = self.parent.projProc.info
        name = self.name
        if not name in info['box_names']:
            return

        hist = info['hists'][name]
        fit_results = info['fit_results']
        subtracted_hists = info['subtracted_hists']
        all_baselines = info['baselines']
        all_centroids = info['centroids']
        all_widths = info['widths']
        all_peaks = info['moved_peaks']
        bgsubs = info['bgsubs']
        hull_ranges = info['hull_ranges']

        ax = self.graphAxes1
        ax.cla()

        ax2 = self.graphAxes2
        ax2.cla()

        if self.histChkBx.isChecked():
            ax.plot(hist, color='k')

        if name in fit_results:
            xs = np.arange(0, len(hist))
            model = info['fit_results'][name]
            convex_hull = hist - info['hists2'][name]

            if self.subHistChkBx.isChecked() and name in subtracted_hists:
                ax2.plot(subtracted_hists[name], color='k')

            if self.fitmodelChkBx.isChecked():
                model_hist = layerlineModel(xs, **model)
                subtracted_model = model_hist-layerlineModelBackground(xs, **model)

                if bgsubs[name] == 1:
                    # Add convexhull background
                    model_hist = model_hist + convex_hull

                ax.plot(model_hist, color='g')
                ax2.plot(subtracted_model, color='g')

            if self.bgChkBx.isChecked():
                if bgsubs[name] == 1:
                    # Add convex hull background
                    ax.fill_between(xs, min(convex_hull), convex_hull, facecolor='b', alpha=0.3)
                else:
                    # Add 3 Gaussians
                    background = layerlineBackground(xs, **model)
                    meridian_bg = meridianBackground(xs, **model) + layerlineBackground(xs, **model)
                    meridian = layerlineModelBackground(xs, **model)
                    ax.fill_between(xs, meridian, meridian_bg, facecolor='r', alpha=0.3)
                    ax.fill_between(xs, meridian_bg, background, facecolor='y', alpha=0.3)
                    ax.fill_between(xs, 0, background, facecolor='b', alpha=0.3)

            if self.centerChkBx.isChecked():
                # Add center line
                ax.axvline(model['centerX'], color='c', alpha=0.7)
                ax2.axvline(model['centerX'], color='c', alpha=0.7)

            if self.peaksChkBx.isChecked():
                # display model peaks
                i = 0
                while 'p_' + str(i) in model:
                    p = model['p_' + str(i)]
                    i += 1
                    ax.axvline(model['centerX'] - p, color='r', alpha=0.7)
                    ax.axvline(model['centerX'] + p, color='r', alpha=0.7)

            if self.maxPeaksChkBx.isChecked():
                peaks = all_peaks[name]
                for p in peaks:
                    d = p - model['centerX']
                    ax2.axvline(model['centerX'] - d, color='b', alpha=0.7)
                    ax2.axvline(model['centerX'] + d, color='b', alpha=0.7)
            # max intensity locations
            # if name in all_peaks:
            #     peaks = all_peaks[name]
            #     for p in peaks:
            #         ax2.axvline(p, color='r', alpha=0.7)

            if name in all_centroids and name in all_baselines and (self.centroidChkBx.isChecked() or self.baselineChkBx.isChecked()):
                # Add baselines and centroids
                centroids = all_centroids[name]
                baselines = all_baselines[name]
                widths = all_widths[name]
                centerX = model['centerX']
                i = 0
                while 'p_' + str(i) in model:
                    c = centroids[i]
                    b = baselines[i]
                    w = widths[i]
                    if self.centroidChkBx.isChecked():
                        ax2.axvline(centerX + c, color='#ff4732')
                        ax2.axvline(centerX - c, color='#ff4732')
                    if self.baselineChkBx.isChecked():
                        ax2.plot(((centerX + c) - w, (centerX + c) + w), (b, b), color='y')
                        ax2.plot(((centerX - c) - w, (centerX - c) + w), (b, b), color='y')

                    i += 1

        if self.hullRangeChkBx.isChecked() and bgsubs[name] == 1 and name in hull_ranges:
            # Color area OUTSIDE convex hull range
            centerX = self.getCenterX()

            ax.axvspan(centerX - hull_ranges[name][0], centerX + hull_ranges[name][0], alpha=0.5, color='k')
            ax.axvspan(0, centerX - hull_ranges[name][1], alpha=0.5, color='k')
            ax.axvspan(centerX + hull_ranges[name][1], len(hist), alpha=0.5, color='k')


            # Update spin box
            self.startHull.setValue(hull_ranges[name][0])
            self.endHull.setValue(hull_ranges[name][1])


        if self.zoom1 is not None:
            ax.set_xlim(self.zoom1[0])
            ax.set_ylim(self.zoom1[1])
        else:
            ax.set_xlim((0, len(hist)))

        if self.zoom2 is not None:
            ax2.set_xlim(self.zoom2[0])
            ax2.set_ylim(self.zoom2[1])
        else:
            ax2.set_xlim((0, len(hist)))

        if self.graphMaxBound is None:
            self.graphMaxBound = [ax.get_xlim(), ax.get_ylim()]

        self.graphFigure1.tight_layout()
        self.graphCanvas1.draw()
        self.graphFigure2.tight_layout()
        self.graphCanvas2.draw()


        # Update Table
        if name in all_centroids and name in all_baselines:
            centroids = all_centroids[name]
            baselines = all_baselines[name]
            widths = all_widths[name]
            nPeaks = len(centroids)
            fit_result = fit_results[name]
            peaks = all_peaks[name] - fit_result['centerX']
            self.resultTable1.setRowCount(nPeaks)
            self.resultTable2.setRowCount(nPeaks)

            for i in range(nPeaks):
                center = fit_result['p_'+str(i)]
                sigma = fit_result['sigma'+str(i)]
                area = fit_result['amplitude' + str(i)]

                item = QTableWidgetItem(str(peaks[i]))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable1.setItem(i, 0, item)

                item = QTableWidgetItem(str(center))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable1.setItem(i, 1, item)

                item = QTableWidgetItem(str(sigma))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable1.setItem(i, 2, item)

                item = QTableWidgetItem(str(area))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable1.setItem(i, 3, item)

                centroid = centroids[i]
                baseline = baselines[i]
                width = widths[i]

                item = QTableWidgetItem(str(baseline))
                self.resultTable2.setItem(i, 0, item)

                item = QTableWidgetItem(str(centroid))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable2.setItem(i, 1, item)

                item = QTableWidgetItem(str(width))
                item.setFlags(Qt.ItemIsEnabled)
                self.resultTable2.setItem(i, 2, item)

        self.need_update = False
        self.syncUI = False



__author__ = 'Jiranun.J'

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy

ext_modules=[ Extension("QF_utilities",
              ["QF_utilities.pyx"],
              libraries=["m"],
              extra_compile_args = ["-ffast-math"])]

setup(
    name='Utilities for Quadrant-Folding',
    ext_modules=ext_modules,
    cmdclass = {"build_ext": build_ext},
    include_dirs=[numpy.get_include()]
)

#
# from distutils.core import setup
# from Cython.Build import cythonize
# import numpy
#
# setup(
#     name='Utilities for Quadrant-Folding',
#     ext_modules=cythonize("QF_utilities.pyx"),
#     include_dirs=[numpy.get_include()]
# )



"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""

__author__ = 'Jiranun.J'

from .pyqt_utils import *
from os.path import exists, join
from threading import Thread
from musclex.utils.file_manager import getFilesAndHdf, createFolder
import fabio
from musclex.utils.image_processor import averageImages
import musclex

class ImageMergerGUI(QMainWindow):
    """
    A class for GUI of Image Merger
    """

    def __init__(self):
        """
        Initial window
        """
        QWidget.__init__(self)
        self.img_list = []
        self.img_grps = []
        self.initUI()
        self.setConnections()

    def initUI(self):
        """
        initial all widgets
        """
        self.setWindowTitle("Muscle X Image Merger v." + musclex.__version__)
        self.centralWid = QWidget(self)
        self.setCentralWidget(self.centralWid)
        self.mainLayout = QGridLayout(self.centralWid)

        self.in_directory = QLineEdit()
        self.in_directory.setEnabled(False)
        self.select_in_folder = QPushButton("Browse")
        self.out_directory = QLineEdit()
        self.select_out_folder = QPushButton("Browse")

        self.frame_number = QSpinBox()
        self.frame_number.setRange(1, 30000)
        self.frame_number.setValue(3)

        self.detailGrp = QGroupBox("Logs")
        self.detailLayout = QVBoxLayout(self.detailGrp)
        self.detail = QPlainTextEdit()
        self.detail.setReadOnly(True)
        self.progressbar = QProgressBar()
        self.progressbar.setHidden(True)
        self.detailLayout.addWidget(self.detail)
        self.detailLayout.addWidget(self.progressbar)

        self.start_button = QPushButton("Start")

        self.mainLayout.addWidget(QLabel("Input Directory : "), 0, 0, 1, 1)
        self.mainLayout.addWidget(self.in_directory, 0, 1, 1, 1)
        self.mainLayout.addWidget(self.select_in_folder, 0, 2, 1, 1)

        self.mainLayout.addWidget(QLabel("Output Directory : "), 1, 0, 1, 1)
        self.mainLayout.addWidget(self.out_directory, 1, 1, 1, 1)
        self.mainLayout.addWidget(self.select_out_folder, 1, 2, 1, 1)

        self.mainLayout.addWidget(QLabel("Number of frames to average : "), 2, 0, 1, 2, Qt.AlignRight)
        self.mainLayout.addWidget(self.frame_number, 2, 2, 1, 1)
        self.mainLayout.addWidget(self.detailGrp, 3, 0, 1, 3)
        self.mainLayout.addWidget(self.start_button, 4, 0, 1, 3, Qt.AlignCenter)

        self.mainLayout.columnStretch(1)
        self.mainLayout.rowStretch(3)
        self.resize(800, 300)
        self.show()

    def setConnections(self):
        """
        Set handler for all widgets
        """
        self.select_in_folder.clicked.connect(self.browse_input)
        self.select_out_folder.clicked.connect(self.browse_output)
        self.frame_number.valueChanged.connect(self.updateImageGroups)
        self.start_button.clicked.connect(self.start_clicked)

    def browse_input(self):
        """
        Handle when Browse for input folder is clicked
        :return:
        """
        path = getAFolder()
        if len(path) > 0:
            self.in_directory.setText(path)
            self.out_directory.setText(join(path, 'merged_results'))
            self.preprocessfolder()

    def browse_output(self):
        """
        Handle when Browse for output folder is clicked
        :return:
        """
        path = getAFolder()
        if len(path) > 0:
            self.out_directory.setText(path)

    def preprocessfolder(self):
        """
        Get all image names in the selected folder
        """
        imgs, _ = getFilesAndHdf(str(self.in_directory.text()))
        self.img_list = sorted(imgs)
        self.updateImageGroups()


    def updateImageGroups(self):
        """
        Group images and update details
        """
        self.img_grps = self.splitImages()
        grps = self.img_grps
        self.detail.clear()
        detail = "Available Groups : \n"
        if len(grps) >= 1:
            for i in range(len(grps)):
                detail += "Group "+ str(i+1)+ " : " + str(grps[i][0]) + " ... " + str(grps[i][-1]) + '\n'

        self.detail.insertPlainText(detail)
        self.detail.moveCursor(QTextCursor.End)


    def splitImages(self):
        """
        Split images into groups
        :return: list of group
        """
        imgs = self.img_list
        frames = self.frame_number.value()
        grps = []
        for i in range(0, len(imgs), frames):
            grps.append(imgs[i:i + frames])

        return grps


    def start_clicked(self):
        """
        handle when Start is clicked
        :return:
        """
        if len(self.img_grps) > 0:
            if self.start_button.text() == 'Start':
                createFolder(str(self.out_directory.text()))
                self.processFolder()
        else:
            errMsg = QMessageBox()
            errMsg.setText('Number of images : 0')
            errMsg.setInformativeText('Please select another folder')
            errMsg.setStandardButtons(QMessageBox.Ok)
            errMsg.setIcon(QMessageBox.Abort)
            errMsg.exec_()

    def processFolder(self):
        """
        merging images
        :return:
        """
        input = str(self.in_directory.text())
        output = str(self.out_directory.text())
        self.progressbar.setHidden(False)
        self.progressbar.setRange(0, len(self.img_grps))
        self.detail.insertPlainText("\n\n--------------- Start ----------------\n\n")
        for i, imgs in enumerate(self.img_grps):
            if len(imgs) > 0:
                self.progressbar.setValue(i)
                first = imgs[0]
                last = imgs[-1]
                f_ind1 = first.rfind('_')
                f_ind2 = first.rfind('.')
                l_ind1 = last.rfind('_')
                l_ind2 = last.rfind('.')

                if f_ind1 == -1 or f_ind2 == -1 or l_ind1 == -1 or l_ind2 == -1 or first[:f_ind1] != last[:l_ind1]:
                    filename = "Group"+str(i+1)+'.tif'
                else:
                    filename = first[:f_ind1] + "_"
                    filename += first[f_ind1 + 1:f_ind2]
                    filename += "_"
                    filename += last[l_ind1 + 1:l_ind2]
                    filename += '.tif'

                # Update details
                details = "- Merging Group "+ str(i+1)+" : \n"
                for j in range(len(imgs)):
                    details += str(j+1)+". "+imgs[j]+"\n"
                details += "To : " + filename +'\n\n'

                self.detail.insertPlainText(details)
                self.detail.moveCursor(QTextCursor.End)
                QApplication.processEvents()
                full_imgs = list(map(lambda f: join(input, f), imgs))
                avg = averageImages(full_imgs)
                fabio.tifimage.tifimage(data=avg).write(join(output, filename))
            else:
                break

        self.detail.insertPlainText("Done. All result images have been saved to "+output)
        self.progressbar.setHidden(True)


"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""
import matplotlib.pyplot as plt
import fabio
import pyFAI
from lmfit import Parameters
from lmfit.models import VoigtModel
from os import makedirs
from os.path import isfile, exists
import pickle
from ..utils.file_manager import fullPath
from ..utils.image_processor import *
from ..utils.histogram_processor import *
import musclex

class DiffractionCentroids():
    """
    A class for Diffraction Centroids processing - go to process() to see all processing steps
    """
    def __init__(self, dir_path, imgList, grp_number, fixRanges, off_mer):
        """
        Initial value for DiffractionCentroids object
        :param dir_path: directory path of input images (str)
        :param imgList: all images in a group (list)
        :param grp_number: image group number - used for writing csv (int)
        :param fixRanges: fixed peak ranges configured my users
        :param off_mer: configuration of off-meridian peaks configured my users
        """
        self.avgImg = self.mergeImages(dir_path, imgList) # avarage all image in a group
        self.mask_thres = getMaskThreshold(self.avgImg)
        self.dir_path = dir_path
        self.version = musclex.__version__
        self.info = {
            "reject" : {"top":[], "bottom": [] }
        }
        self.cachefile, cinfo = self.loadCache(dir_path, imgList)
        self.info.update(cinfo)
        self.info["filelist"] = imgList
        self.info["grp_num"] = grp_number+1
        self.fixRanges = fixRanges
        self.init_off_mer = off_mer

    def mergeImages(self, dir_path, imgList):
        """
        Merge all image in imgList
        :param dir_path: directory of images
        :param imgList: all merging images
        :return:
        """
        imgList2 = []
        for fname in imgList:
            fullname = fullPath(dir_path, fname)
            img = fabio.open(fullname).data
            imgList2.append(img)
        return np.mean(imgList2, axis=0)

    def loadCache(self, dir_path, imgList):
        """
        Load info dict from cache. Cache file will be filename.info in folder "qf_cache"
        :param dir_path: directory of images
        :param imgList: input images
        :return: cached info (dict)
        """
        cache_path = fullPath(dir_path, 'dc_cache')
        cache_filename = imgList[0]+'_'+ imgList[-1]+ '.info'
        cachefile = fullPath(cache_path, cache_filename)
        info = {}

        if isfile(cachefile):
            cinfo = pickle.load(open(cachefile, "rb"))
            if cinfo != None:
                if cinfo['program_version'] == self.version:
                    info = cinfo

        return cachefile, info

    def process(self, flags):
        """
        All processing steps
        """
        imgList = self.info['filelist']
        print(str(imgList[0])+' ... '+str(imgList[-1])+' are being processed...')
        self.updateInfo(flags)
        self.findCenter()
        self.findRotationAngle()
        self.calculateRmin()
        self.getIntegrateArea()
        self.setConvexhullPoints()
        self.getHistograms()
        self.getPeaks()
        self.correctPeaks()
        self.calculateBaselines()
        self.calculateCentroids()

        if self.init_off_mer is not None:
            self.getOffMeridianRanges()
            self.getOffMerRminmax()
            self.getOffMeridianHistograms()
            self.getOffMeridianPeaks()
            self.getOffMeridianBaselines()
            self.getOffMeridianInfos()

        if "no_cache" not in self.info:
            self.cacheInfo()

    def removeInfo(self, k):
        """
        Remove k from info dict
        :param k: key of dictionary
        :return: -
        """
        if k == 'hists':
            for ky in ['top_hist', 'top_hull', 'bottom_hist', 'bottom_hull']:
                self.removeInfo(ky)
        if k in self.info:
            del self.info[k]

    def updateInfo(self, flags):
        if flags['orientation_model'] is None:
            if 'orientation_model' not in self.info:
                flags['orientation_model'] = 0
            else:
                del flags['orientation_model']
        self.info.update(flags)

    def findCenter(self):
        """
        Find center of diffraction, and keep it in self.info["center"]
        This calculation will affect rotation angle, so self.info["rotationAngle"] will be removed
        """
        if 'center' in self.info:
            return
        self.info['center'] = getCenter(self.avgImg)
        print("center = "+str(self.info['center']))
        self.removeInfo('rotationAngle')

    def findRotationAngle(self):
        """
        Find rotation angle of diffraction, and keep it in self.info["rotationAngle"]
        This calculation will affect R-min, so self.info["rmin"] will be removed
        """
        if 'rotationAngle' in self.info:
            return
        center = self.info['center']
        self.info['rotationAngle'] = getRotationAngle(self.avgImg, center, self.info['orientation_model'])
        print("rotation angle = " + str(self.info['rotationAngle']))
        self.removeInfo('rmin')

    def calculateRmin(self):
        """
        Find R-min of diffraction, and keep it in self.info["rmin"]
        This calculation will affect integrated area (meridian), so self.info["int_area"] will be removed
        """
        if 'rmin' in self.info:
            return
        img = copy.copy(self.avgImg)
        center = self.info['center']

        if img.shape == (1043, 981):
            det = "pilatus1m"
        else:
            det = "agilent_titan"

        corners = [(0, 0), (img.shape[1], 0), (0, img.shape[0]), (img.shape[1], img.shape[0])]
        npt_rad = int(round(max([distance(center, c) for c in corners])))
        ai = pyFAI.AzimuthalIntegrator(detector=det)
        ai.setFit2D(100, center[0], center[1])
        tth, I = ai.integrate1d(img, npt_rad, unit="r_mm")
        # tth, I = ai.integrate1d(img, npt_rad, unit="r_mm")
        self.info['rmin'] = getFirstVallay(I)
        print("R-min = "+str(self.info['rmin']))
        self.removeInfo('int_area')

    def getIntegrateArea(self):
        """
        Find intergrated area or meridian lines of diffraction, and keep it in self.info["int_area"]
        This calculation will affect start and end points of convexh hull applying points, so self.info["top_se"] and self.info["bottom_se"] will be removed
        """
        if 'int_area' in self.info:
            return
        center = self.info['center']
        rmin = self.info['rmin']
        img = getCenterRemovedImage(copy.copy(self.avgImg), self.info['center'], self.info['rmin']) # remove center location
        rotate_img = self.getRotatedImage(img) # rotate image
        l = max(0,center[0]-rmin) # initial guess for left line
        r = min(center[0]+rmin, rotate_img.shape[1]) # initial guess for right line
        area = rotate_img[:,l:r] # Get the area by initial guess
        hist = np.sum(area, axis=0) # find verital histogram
        hull = convexHull(hist)  # appl convex hull to the histogram

        if len(hist) > 0:
            max_loc = np.argmax(hull)
            r = 1
            l = 1
            # Find start and end points by finding first zero value from on both side
            while max_loc-l >= 0 and max_loc+r < len(hull) and (hull[max_loc-l] != 0 or hull[max_loc+r] != 0):
                if hull[max_loc-l] != 0:
                    l += 1
                if hull[max_loc+r] != 0:
                    r += 1

            # Assume that meridian size should not smaller than 50% or R-min
            if max_loc+r < center[0] or max_loc-l > center[0] or abs(r+l) < rmin*.5:
                self.info['int_area'] = (int(round(center[0] - rmin*.5)), int(round(center[0] + rmin*.5)) + 1)
            else:
                center = center[0]-rmin+max_loc
                self.info['int_area'] = (int(round(center-l*1.2)), int(round(center+r*1.2))+1)
        else:
            self.info['int_area'] = (int(round(center[0] - rmin * .5)), int(round(center[0] + rmin * .5)) + 1)
        print("integrated area = "+str(self.info['int_area']))
        self.removeInfo('top_se')
        self.removeInfo('bottom_se')

    def getRotatedImage(self, img = None, angle = None):
        """
        Get rotated image by angle. If the input params are not specified. image = original input image, angle = self.info["rotationAngle"]
        :param img: input image
        :param angle: rotation angle
        :return: rotated image
        """
        if img is None:
            img = copy.copy(self.avgImg)
        if angle is None:
            angle = self.info['rotationAngle']
        if '90rotation' in self.info and self.info['90rotation'] is True:
            angle = angle - 90 if angle > 90 else angle + 90

        return rotateImage(img, self.info["center"], angle, self.mask_thres)

    def setConvexhullPoints(self):
        """
        Set start and and points for convex hull in both top and bottom side. Init by start = R-min and end = 0.7*half of image size.
        This values will be kept in self.info["top_se"] and self.info["bottom_se"]
        This calculation will affect histograms, so all histograms will be removed from self.info
        """
        if 'top_se' not in self.info:
            if 'top_fixed_se' in self.info:
                self.info['top_se'] = self.info['top_fixed_se']
            else:
                self.info['top_se'] = (self.info['rmin'], int(round(min(self.avgImg.shape[0] / 2, self.avgImg.shape[1] / 2) * 0.7)))
            self.removeInfo('top_hist')
            self.removeInfo('top_hull')
        if 'bottom_se' not in self.info:
            if 'bottom_fixed_se' in self.info:
                self.info['bottom_se'] = self.info['bottom_fixed_se']
            else:
                self.info['bottom_se'] = (self.info['rmin'], int(round(min(self.avgImg.shape[0]/2, self.avgImg.shape[1]/2) * 0.7)))
            self.removeInfo('bottom_hist')
            self.removeInfo('bottom_hull')

    def getHistograms(self):
        """
        Get original histograms and background subtracted histogram ( subtracted by convex hull )
        The original histogram will be kept in self.info["top_hist"] and self.info["bottom_hist"]
        The background subtracted histogram will be kept in self.info["top_hull"] and self.info["bottom_hull"]
        These changing will affect peak locations, so peaks will be removed from self.info
        """
        if 'top_hist' in self.info and 'top_hull' in self.info and 'bottom_hist' in self.info and 'bottom_hull' in self.info:
            return

        center_y = self.info['center'][1]
        int_area = self.info['int_area']
        img = self.getRotatedImage(copy.copy(self.avgImg), self.info['rotationAngle'])
        img_area = img[:,int_area[0]: int_area[1]]
        ignore = np.array([any(line <= self.mask_thres) for line in img_area])
        hist = np.sum(img_area, axis=1)

        top_hist, top_ignore, bottom_hist, bottom_ignore = self.splitHist(center_y, hist, ignore)

        if not ('top_hist' in self.info or 'top_hull' in self.info):
            top_hull = convexHull(top_hist, start_p=self.info['top_se'][0], end_p=self.info['top_se'][1], ignore=top_ignore)
            self.info['top_hist'] = np.array(top_hist)
            self.info['top_hull'] = np.array(top_hull)
            self.removeInfo('pre_top_peaks')

        if not ('bottom_hist' in self.info or 'bottom_hull' in self.info):
            bottom_hull = convexHull(bottom_hist, start_p=self.info['bottom_se'][0], end_p=self.info['bottom_se'][1], ignore=bottom_ignore)
            self.info['bottom_hist'] = np.array(bottom_hist)
            self.info['bottom_hull'] = np.array(bottom_hull)
            self.removeInfo('pre_bottom_peaks')

    def getPeaks(self):
        """
        Get pre peaks from histograms with specified fixed ranges or without. These peaks won't be used until it's correct.
        These pre peaks will be kept in self.info["pre_[side]_peaks"]
        This calculation will affect reak peaks, so peaks will be removed from self.info
        """
        if 'pre_top_peaks' not in self.info:
            if len(self.fixRanges) > 0:
                self.info['pre_top_peaks'] = self.getPeaksFromRanges(self.info['top_hull'], self.fixRanges)
            else:
                self.info['pre_top_peaks'] = getPeaksFromHist(self.info['top_hull'])
            self.removeInfo('top_peaks')

        if 'pre_bottom_peaks' not in self.info:
            if len(self.fixRanges) > 0:
                self.info['pre_bottom_peaks'] = self.getPeaksFromRanges(self.info['bottom_hull'], self.fixRanges)
            else:
                self.info['pre_bottom_peaks'] = getPeaksFromHist(self.info['bottom_hull'])
            self.removeInfo('bottom_peaks')

    def getPeaksFromRanges(self, hist, fix_ranges):
        """
        Get Peaks from specified peak ranges
        :param hist: background subtracted histogram
        :param fix_ranges: fixed ranges specified by users
        :return: -
        """
        results = []
        for fr in fix_ranges:
            r = fr[1]
            start = min(r[0], len(hist)-2)
            end = min(r[1], len(hist)-1)
            peak = start + np.argmax(hist[start:end + 1])
            results.append(peak)
        return results

    def correctPeaks(self):
        """
        Correct pre-peak locatons by moving them to the local maximum point.
        These result peaks are considered as real peaks. They will be kept in self.info["[side]_peaks"]
        Peaks location will affect baselines, so baselines will be removed from self.info
        :return:
        """
        if 'top_peaks' not in self.info:
            if len(self.fixRanges) == 0:
                moved_peaks = self.movePeaks(self.info['top_hull'], self.info['pre_top_peaks'])
                self.info['top_peaks'] = moved_peaks
                self.info['top_names'] = [str(p) for p in moved_peaks]
            else:
                self.info['top_peaks'] = self.info['pre_top_peaks']
                self.info['top_names'] = [self.fixRanges[i][0] for i in range(len(self.fixRanges))]
            self.removeInfo('top_baselines')

        print("Top peaks : ")
        for i in range(len(self.info['top_peaks'])):
            print(str(self.info['top_names'][i])+" : "+str(self.info['top_peaks'][i]))

        if 'bottom_peaks' not in self.info:
            if len(self.fixRanges) == 0:
                moved_peaks = self.movePeaks(self.info['bottom_hull'], self.info['pre_bottom_peaks'])
                self.info['bottom_peaks'] = moved_peaks
                self.info['bottom_names'] = [str(p) for p in moved_peaks]
            else:
                self.info['bottom_peaks'] = self.info['pre_bottom_peaks']
                names = [self.fixRanges[i][0] for i in range(len(self.fixRanges))]
                self.info['bottom_names'] = names
            self.removeInfo('bottom_baselines')
        print("Bottom peaks : ")
        for i in range(len(self.info['bottom_peaks'])):
            print(str(self.info['bottom_names'][i]) + " : " + str(self.info['bottom_peaks'][i]))

    def movePeaks(self, hist, peaks, dist = 10):
        """
        Move peaks to their local maximum. Duplicated peak locations will be removed
        :param hist: input histogram
        :param peaks: approximate peak locations
        :param dist: maximum distance of local maximum
        :return: sorted moved peaks
        """
        peakList = []
        smooth_hist = smooth(hist)
        for p in peaks:
            new_peak = p
            while True:
                start = max(0, p-dist)
                end = min(len(hist), p+dist)
                new_peak = start + np.argmax(hist[start:end])

                if abs(p-new_peak) < 4:
                    break
                else:
                    left = min(p, new_peak)
                    right = max(p, new_peak)
                    if all(smooth_hist[left+1:right] > p):
                        break
                dist = dist/2
            peakList.append(new_peak)
        return sorted(list(set(peakList)))

    def calculateBaselines(self):
        """
        Find baselines of peaks. Initial with half-height of peaks
        Baselines will be kept in self..info["[side]_baselines"].
        This calulation might affact other infos : centroids width and intensity
        """
        if 'top_baselines' not in self.info:
            hist = self.info['top_hull']
            peaks = self.info['top_peaks']
            self.info['top_baselines'] = [hist[p] / 2. for p in peaks]
            self.removeInfo('top_centroids')
        print("Top baselines = "+str(self.info['top_baselines']))
        if 'bottom_baselines' not in self.info:
            hist = self.info['bottom_hull']
            peaks = self.info['bottom_peaks']
            self.info['bottom_baselines'] = [hist[p]/2. for p in peaks]
            self.removeInfo('bottom_centroids')
        print("Bottom baselines = "+str(self.info['bottom_baselines']))

    def calculateCentroids(self):
        """
        Calculate all other peaks infomation including centroid, width, and intensity(area)
        This results will be kept in self.info
        """
        if 'top_centroids' not in self.info:
            hist = self.info['top_hull']
            peaks = self.info['top_peaks']
            baselines = self.info["top_baselines"]
            results = getPeakInformations(hist, peaks, baselines)
            self.info['top_centroids'] = results['centroids']
            self.info['top_widths'] = results['widths']
            self.info['top_areas'] = results['areas']
        print("Top centroids = "+ str(self.info['top_centroids']))
        if 'bottom_centroids' not in self.info:
            hist = self.info['bottom_hull']
            peaks = self.info['bottom_peaks']
            baselines = self.info["bottom_baselines"]
            results = getPeakInformations(hist, peaks, baselines)
            self.info['bottom_centroids'] = results['centroids']
            self.info['bottom_widths'] = results['widths']
            self.info['bottom_areas'] = results['areas']
        print("Bottom centroids = " + str(self.info['bottom_centroids']))

    def fitModel(self, hist, peaks, baselines):
        # JUST FOR TEST
        # Fit Voigt model to histogram using peaks and baselines
        # Currently, this function is not used by any process
        new_hist = np.zeros(len(hist))
        pars = Parameters()
        mean_margin = 3

        for i in range(len(peaks)):
            p = peaks[i]
            baseline = baselines[i]
            width, _ = getWidth(hist, p, baseline)
            new_hist[p - width:p + width] = hist[p - width:p + width]
            prefix = "v"+str(i + 1) + '_'
            init_amp = hist[p] * width * np.sqrt(2 * np.pi)
            voigt = VoigtModel(prefix=prefix)
            pars.add(prefix + 'center', p,  min=p - mean_margin, max=p + mean_margin)
            pars.add(prefix + 'sigma',width, min = 0, max = width*3.)
            pars.add(prefix + 'amplitude', init_amp, min = 0, max = init_amp*3.)
            pars.add(prefix + 'gamma', width,  min = 0, max = width*3.)
            pars.add(prefix + 'fwhm', baseline, min=0, max=baseline * 3.)
            pars.add(prefix + 'height',  hist[p], min=0, max= hist[p] * 3.)
            if i == 0:
                model = voigt
            else:
                model += voigt

        xs = np.arange(0, len(new_hist))
        result = model.fit(np.array(new_hist), params = pars, x = xs).values
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(hist)
        ax.plot(new_hist)
        ax.plot(model.eval(x=xs, **result))
        fig.show()
        print(str(result))

    def setBaseline(self, side, peak_num, new_baseline):
        """
        Set new baselines of meridina peaks by users. Remove centroids from self.info as it needs to be re-calculated.
        :param side: "top" or "bottom"
        :param peak_num: peak number from left to right (int)
        :param new_baseline: new baseline value or percent of peak height (str or float)
        """
        new_baseline = str(new_baseline)
        hist = self.info[side + "_hull"]
        peaks = self.info[side+ "_peaks"]
        baselines = self.info[side + "_baselines"]
        height = hist[peaks[peak_num]]
        if "%" in new_baseline:
            # if new_baseline contain "%", baseline value will use this as percent of peak height
            percent = float(new_baseline.rstrip("%"))
            baseline = height*percent/100.
        elif len(new_baseline) == 0:
            # if new_baseline is empty, baseline will by half-height
            baseline = float(height*.5)
        else:
            baseline = float(new_baseline)

        if height > baseline:
            baselines[peak_num] = baseline

        self.removeInfo(side+"_centroids")

    def setOffMerBaseline(self, quadrant, ind, new_baseline):
        """
        Set new baselines of off-meridian peaks by users. Remove other infos from self.info as it needs to be re-calculated.
        :param quadrant: "top_left", "top_right, "bottom_left", or "bottom_right"
        :param ind: peak number from left to right (int)
        :param new_baseline: new baseline value or percent of peak height (str or float)
        """
        new_baseline = str(new_baseline)
        hist = self.info["off_mer_hists"]["hulls"][quadrant]
        peak = self.info["off_mer_peaks"][quadrant][ind]
        baselines = self.info["off_mer_baselines"][quadrant]
        height = hist[peak]

        if "%" in new_baseline:
            # if new_baseline contain "%", baseline value will use this as percent of peak height
            percent = float(new_baseline.rstrip("%"))
            baseline = height * percent / 100.
        elif len(new_baseline) == 0:
            # if new_baseline is empty, baseline will by half-height
            baseline = float(height * .5)
        else:
            baseline = float(new_baseline)

        if height > baseline:
            baselines[ind] = baseline

        self.removeInfo("off_mer_peak_info")

    def getOffMeridianRanges(self):
        """
        Get off-meridian ranges as position in the image from x1,x2,x3,x4 which are specified by users
        """
        if "x1" not in self.info:
            centerX = self.info["center"][0]
            self.info["x1"] = centerX - self.init_off_mer["x1"]
            self.info["x2"] = centerX - self.init_off_mer["x2"]
            self.info["x3"] = centerX + self.init_off_mer["x3"]
            self.info["x4"] = centerX + self.init_off_mer["x4"]
            self.removeInfo("off_mer_rmin_rmax")

    def splitHist(self, center_y, hist, ignore):
        """
        Split histogram to top and bottom by center_y
        :param center_y: center location of histogram (int)
        :param hist: input histogram (list)
        :param ignore: ignored locations in histogram (list of boolean)
        :return: top_hist, top_ignore, bottom_hist, bottom_ignore (list)
        """
        top_hist = hist[center_y:]
        top_ignore = ignore[center_y:]
        bottom_hist = hist[:center_y]
        bottom_hist = bottom_hist[::-1]
        bottom_ignore = ignore[:center_y]
        bottom_ignore = bottom_ignore[::-1]
        return top_hist, top_ignore, bottom_hist, bottom_ignore

    def getOffMerRminmax(self):
        if 'off_mer_rmin_rmax' not in self.info:
            if 'fixed_offmer_hull_range' in self.info:
                self.info['off_mer_rmin_rmax'] = self.info['fixed_offmer_hull_range']
            else:
                rmin, rmax = self.initOffMeridianPeakRange()
                self.info['off_mer_rmin_rmax'] = (rmin, rmax)
            self.removeInfo("off_mer_hists")

    def getOffMeridianHistograms(self):
        """
        Produce histograms of off-meridian ranges.
        All histograms will be kept in self.info["off_mer_hists"]["hists"]
        All backgound subtracted histograms will be kept in self.info["off_mer_hists"]["hull"]
        """
        if "off_mer_hists" not in self.info:
            center_y = self.info["center"][1]
            x1 = self.info["x1"]
            x2 = self.info["x2"]
            x3 = self.info["x3"]
            x4 = self.info["x4"]
            img = self.getRotatedImage()
            left_area = img[:, x2:x1]
            right_area = img[:, x3:x4]
            # left_ignore = np.array([any(line < 0) for line in left_area])
            # right_ignore = np.array([any(line < 0) for line in right_area])

            left_ignore = np.array([sum(np.array(line) < 0) > 0.1*len(line) for line in left_area])
            right_ignore = np.array([sum(np.array(line) < 0) > 0.1*len(line) for line in right_area])
            left_hist = np.sum(left_area, axis=1)
            right_hist = np.sum(right_area, axis=1)
            top_left_hist, top_left_ignore, bottom_left_hist, bottom_left_ignore = self.splitHist(center_y, left_hist,
                                                                                                  left_ignore)
            top_right_hist, top_right_ignore, bottom_right_hist, bottom_right_ignore = self.splitHist(center_y, right_hist,
                                                                                                  right_ignore)
            start, end = self.info['off_mer_rmin_rmax']
            top_left_hull = convexHull(top_left_hist, start_p=start, end_p=end, ignore=top_left_ignore)
            bottom_left_hull = convexHull(bottom_left_hist, start_p=start, end_p=end, ignore=bottom_left_ignore)
            top_right_hull = convexHull(top_right_hist, start_p=start, end_p=end, ignore=top_right_ignore)
            bottom_right_hull = convexHull(bottom_right_hist, start_p=start, end_p=end, ignore=bottom_right_ignore)

            self.info["off_mer_hists"] = {
                "hists" : {
                    "top_left" : np.array(top_left_hist),
                    "top_right": np.array(top_right_hist),
                    "bottom_left" : np.array(bottom_left_hist),
                    "bottom_right": np.array(bottom_right_hist)
                },
                "hulls": {
                    "top_left": np.array(top_left_hull),
                    "top_right": np.array(top_right_hull),
                    "bottom_left": np.array(bottom_left_hull),
                    "bottom_right": np.array(bottom_right_hull)
                }
            }
            self.removeInfo("off_mer_peaks")

    def getOffMeridianPeaks(self):
        """
        Get peak 51 and 59 from 4 quadrants by using peak ranges speicifed by users
        These peaks will be kept in self.info["off_mer_peaks"]
        This might affect baselines, so off_mer_baselines is removed from self.info
        """
        if "off_mer_peaks" not in self.info:
            peaks = {}
            hulls = self.info["off_mer_hists"]["hulls"]
            peak_ranges = [("59",(self.init_off_mer["s59"], self.init_off_mer["e59"])),
                           ("51",(self.init_off_mer["s51"], self.init_off_mer["e51"]))]
            for k in hulls.keys():
                peaks[k] = self.getPeaksFromRanges(hulls[k], peak_ranges)
            self.info["off_mer_peaks"] = peaks
            self.removeInfo("off_mer_baselines")

    def getOffMeridianBaselines(self):
        """
        Get baselines of peak 51 and 59 from 4 quadrants. Init with half-height of peaks
        These baselines will be kept in self.info["off_mer_baselines"]
        This might affect peak infos, so off_mer_peak_info is removed from self.info
        """
        if "off_mer_baselines" not in self.info:
            baselines = {}
            peaks = self.info["off_mer_peaks"]
            hulls = self.info["off_mer_hists"]["hulls"]
            for k in peaks.keys():
                baselines[k] = [hulls[k][p]*.5 for p in peaks[k]]
            self.info["off_mer_baselines"] = baselines
            self.removeInfo("off_mer_peak_info")

    def getOffMeridianInfos(self):
        """
        Get information of peak 51 and 59 from 4 quadrants including centroid, width, intersection with baseline, area (intensity)
        These info will be kept in self.info["off_mer_peak_info"]
        """
        if "off_mer_peak_info" not in self.info:
            all_info = {}
            peaks = self.info["off_mer_peaks"]
            hulls = self.info["off_mer_hists"]["hulls"]
            baselines = self.info["off_mer_baselines"]

            for k in peaks.keys():
                peak_list = peaks[k]
                hull = hulls[k]
                baseline_list = baselines[k]
                results = getPeakInformations(hull, peak_list, baseline_list)
                all_info[k] = copy.copy(results)

            self.info["off_mer_peak_info"] = all_info

    def initOffMeridianPeakRange(self):
        """
        Find start and end points of peak 51 and 59 for applying convex hull
        """
        return int(round(0.9*self.init_off_mer["s59"])), int(round(1.1*self.init_off_mer["e51"]))

    def cacheInfo(self):
        """
        Save info dict to cache. Cache file will be save as filename.info in folder "qf_cache"
        :return: -
        """

        cache_path = fullPath(self.dir_path, 'dc_cache')

        if not exists(cache_path):
            makedirs(cache_path)

        self.info["program_version"] = self.version
        pickle.dump(self.info, open(self.cachefile, "wb"))



"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""


from os import makedirs
import csv
from ..utils.file_manager import *

class DC_CSVManager():
    """
    A class taking care of writing summary file
    """
    def __init__(self, dir_path, nFrames, fix_ranges = []):
        self.nFrames = nFrames # number of frames that average

        sum_path = fullPath(dir_path, "dc_summary")

        if not exists(sum_path):
            makedirs(sum_path)

        if self.nFrames > 0:
            # if number of frames that average is specified, summary fill will be summary_nf.csv while n is number of frames
            self.filename = fullPath(sum_path, "summary_"+str(nFrames)+"f")
        else:
            # otherwise, summary file will be summary_m.csv
            self.filename = fullPath(sum_path, "summary_m")

        fix_range_names = [fr[0] for fr in fix_ranges]
        if len(fix_range_names) > 0:
            for n in fix_range_names:
                self.filename += ("_" + str(n))
        self.filename += ".csv"

        self.allpeaks_data = {} # Collect meridian peaks data
        self.offmer_data = {} # Collect off-meridian peaks data
        self.grp_to_key = {} # Map group number with key
        self.gen_data = {} # general data of Diffraction Centroid. Now, there's only integrated width
        self.headers = [] # header of csv file
        self.keyToImages = {} # Map key to image names
        self.loadSummary() # load exist summary file before adding new data

    def getRow(self, key):
        # return index of key in header
        return self.headers.index(key)

    def loadSummary(self):
        """
        Load exist summary file, and encapsulate them
        """
        self.allpeaks_data = {}
        self.offmer_data = {}
        self.gen_data = {}
        self.keyToImages = {}
        self.grp_to_key = {}
        currentGroup = -1
        header = []
        # if self.nFrames == 0:
        #     # Files are manually selected
        #     if exists(self.filename):
        #         with open(self.filename, "rb") as csvfile:
        #             reader = csv.reader(csvfile)
        #             peak_data = []
        #             images = []
        #             g_data = {}
        #             off_mer = {}
        #
        #             for row in reader:
        #                 if len(row) == 0 :
        #                     continue
        #
        #                 if row[0] == "group":
        #                     if len(images) > 0:
        #                         k = "...".join([images[0], images[-1]])
        #                         self.allpeaks_data[k] = peak_data
        #                         self.offmer_data[k] = off_mer
        #                         self.gen_data[k] = g_data
        #                         self.keyToImages[k] = images
        #                         self.grp_to_key[int(currentGroup)] = k
        #
        #                     header = copy.copy(row)
        #                     peak_data = []
        #                     off_mer = {}
        #                     images = []
        #                     g_data = {}
        #                     continue
        #
        #                 elif row[0] != "":
        #                     currentGroup = row[0]
        #                     images.append(row[1])
        #                     g_data["integration area width"] = row[2]
        #                     for i in range(3, len(row) - 3):
        #                         if "centroid" in header[i]:
        #                             if "average" in header[i]:
        #                                 continue
        #                             head = header[i]
        #                             head = head.split(" ")
        #                             if "_" in head[0] and head[1] == "59":
        #                                 quadrant = head[0]
        #                                 off_mer[quadrant] = {
        #                                     "centroid": [row[header.index(quadrant + " 59 " +"centroid")],
        #                                                  row[header.index(quadrant + " 51 " +"centroid")]],
        #                                     "baseline": [row[header.index(quadrant + " 59 " + "baseline")],
        #                                                  row[header.index(quadrant + " 51 " + "baseline")]],
        #                                     "intensity": [row[header.index(quadrant + " 59 " + "intensity")],
        #                                                  row[header.index(quadrant + " 51 " + "intensity")]],
        #                                 }
        #                             else:
        #                                 side = head[0]
        #                                 name = head[1]
        #                                 new_peak = {
        #                                     "side" : side,
        #                                     "name" : name,
        #                                     "centroid" : row[i],
        #                                     "baseline" : row[header.index(side + " " + name + " " + "baseline")],
        #                                     "intensity" : row[header.index(side + " " + name + " " + "intensity")]
        #                                 }
        #                                 peak_data.append(new_peak)
        #                 else:
        #                     images.append(row[1])
        #
        #             if len(images) > 0:
        #                 k = "...".join([images[0], images[-1]])
        #                 self.allpeaks_data[k] = peak_data
        #                 self.offmer_data[k] = off_mer
        #                 self.gen_data[k] = g_data
        #                 self.keyToImages[k] = images
        #                 self.grp_to_key[int(currentGroup)] = k
        # else:
        # Files are selected by auto-grouping
        if exists(self.filename):
            with open(self.filename, "r") as csvfile:
                reader = csv.reader(csvfile)
                peak_data = []
                images = []
                g_data = {}
                off_mer = {}

                for row in reader:
                    if len(row) == 0:
                        continue

                    if row[0] == "group":
                        header = row[:]
                        continue

                    if len(row[0]) > 0:
                        if len(images) > 0:
                            k = "...".join([images[0], images[-1]])
                            self.allpeaks_data[k] = peak_data
                            self.offmer_data[k] = off_mer
                            self.gen_data[k] = g_data
                            self.keyToImages[k] = images
                            self.grp_to_key[int(currentGroup)] = k
                        peak_data = []
                        images = []
                        g_data = {}
                        off_mer = {}

                    if row[0] != "":
                        currentGroup = row[0]
                        images.append(row[1])
                        g_data["integration area width"] = row[2]
                        for i in range(3, len(row) - 3):
                            if "centroid" in header[i]:
                                if "average" in header[i]:
                                    continue
                                head = header[i]
                                head = head.split(" ")
                                if "_" in head[0] and head[1] == "59":
                                    quadrant = head[0]
                                    off_mer[quadrant] = {
                                        "centroid": [row[header.index(quadrant + " 59 " +"centroid")],
                                                     row[header.index(quadrant + " 51 " +"centroid")]],
                                        "baseline": [row[header.index(quadrant + " 59 " + "baseline")],
                                                     row[header.index(quadrant + " 51 " + "baseline")]],
                                        "intensity": [row[header.index(quadrant + " 59 " + "intensity")],
                                                     row[header.index(quadrant + " 51 " + "intensity")]],
                                    }
                                else:
                                    side = head[0]
                                    name = head[1]
                                    new_peak = {
                                        "side": side,
                                        "name": name,
                                        "centroid": row[i],
                                        "baseline": row[header.index(side + " " + name + " " + "baseline")],
                                        "intensity": row[header.index(side + " " + name + " " + "intensity")]
                                    }
                                    peak_data.append(new_peak)
                    else:
                        images.append(row[1])

                if len(images) > 0:
                    k = "...".join([images[0], images[-1]])
                    self.allpeaks_data[k] = peak_data
                    self.offmer_data[k] = off_mer
                    self.gen_data[k] = g_data
                    self.keyToImages[k] = images
                    self.grp_to_key[int(currentGroup)] = k

    def writeNewData(self, info):
        """
        Add new data (DiffractionCentroids info) to encapsulated member, and rewrite summary
        :param info: DiffractionCentroids info (dict)
        """
        fileList = info["filelist"]
        k = "...".join([fileList[0], fileList[-1]])
        if "summary_m" in self.filename:
            grpnum = len(self.grp_to_key.keys()) + 1
            for (group, key) in self.grp_to_key.items():
                if key == k:
                    grpnum = group
            self.grp_to_key[grpnum] = k
        else:
            self.grp_to_key[info["grp_num"]] = k
        self.keyToImages[k] = fileList

        if k in self.allpeaks_data:
            del self.allpeaks_data[k]

        if k in self.offmer_data:
            del self.offmer_data[k]

        data = []
        for side in ["top", "bottom"]:
            centroids = info[side + "_centroids"]
            peaks = info[side + "_peaks"]
            baselines = info[side + "_baselines"]
            areas = info[side + "_areas"]
            names = info[side + "_names"]
            reject_state = info["reject"][side]
            for i in range(len(peaks)):
                peak_dict = {}
                peak_dict["side"] = side
                peak_dict["name"] = names[i]
                if names[i] in reject_state:
                    peak_dict["centroid"] = "_"+str(centroids[i])
                    peak_dict["peak"] = "_"+str(peaks[i])
                    peak_dict["intensity"] = "_"+str(areas[i])
                    peak_dict["baseline"] = "_"+str(baselines[i])
                else:
                    peak_dict["centroid"] = str(centroids[i])
                    peak_dict["peak"] = str(peaks[i])
                    peak_dict["intensity"] = str(areas[i])
                    peak_dict["baseline"] = str(baselines[i])

                data.append(peak_dict)

        off_mer_data = {}
        if "off_mer_baselines" in info:
            for q in ["top_left", "top_right", "bottom_left", "bottom_right"]:
                dat = {
                    "baseline" : info["off_mer_baselines"][q],
                    "centroid" : info["off_mer_peak_info"][q]["centroids"],
                    "intensity" : info["off_mer_peak_info"][q]["areas"]
                }
                off_mer_data[q] = dat

        self.allpeaks_data[k] = data
        self.offmer_data[k] = off_mer_data
        l,r = info["int_area"]
        self.gen_data[k] = {"integration area width":abs(r-l)}
        self.rewriteSummary()

    def getOffMerHeaders(self):
        """
        Get Off-meridian header for csv file
        """
        header = []
        for t in ["centroid", "baseline", "intensity"]:
            for p in ["59", "51"]:
                for q in ["top_left", "top_right", "bottom_left", "bottom_right"]:
                    header.append(q + " " + p + " "+t)
                header.append(p + " average "+t)
        return header

    def toInt(self, s):
        try:
            result = int(s)
        except Exception:
            result = s
        return result

    def rewriteSummary(self):
        """
        Re-write summary file from encapsulated data
        :return:
        """
        # if self.nFrames == 0:
        #     # If files are selected manually
        #     with open(self.filename, "w") as csvfile:
        #         writer = csv.writer(csvfile, delimiter = ",")
        #         grpKey = sorted(self.grp_to_key.items(), key= lambda (g,k) : int(g))
        #         for (g, k) in grpKey:
        #             peak_data = self.allpeaks_data[k]
        #             g_data = self.gen_data[k]
        #             filelist = self.keyToImages[k]
        #             off_mer = self.offmer_data[k]
        #
        #             ### write header for each group ###
        #             header = ["group", "filename", "integration area width"]
        #             top_peaks = sorted([d for d in peak_data if d["side"] == "top"], key=lambda d: self.toInt(d["name"]))
        #             bottom_peaks = sorted([d for d in peak_data if d["side"] == "bottom"], key=lambda d: self.toInt(d["name"]))
        #             nTop = len(top_peaks)
        #             nBottom = len(bottom_peaks)
        #             min_nPeaks = min(nTop, nBottom)
        #
        #             for i in range(min_nPeaks):
        #                 peak_header = ["top " + top_peaks[i]["name"] + " centroid", "bottom " + bottom_peaks[i]["name"] + " centroid", "average " + bottom_peaks[i]["name"] + " centroid",
        #                                "top " + top_peaks[i]["name"] + " baseline", "bottom " + bottom_peaks[i]["name"] + " baseline", "average " + bottom_peaks[i]["name"] + " baseline",
        #                                "top " + top_peaks[i]["name"] + " intensity", "bottom " + bottom_peaks[i]["name"] + " intensity", "average " + bottom_peaks[i]["name"] + " intensity"]
        #                 header.extend(peak_header)
        #
        #             if nTop > nBottom:
        #                 extra = "top"
        #                 extra_peaks = top_peaks
        #             else:
        #                 extra = "bottom"
        #                 extra_peaks = bottom_peaks
        #
        #             for i in range(max(nTop, nBottom) - min_nPeaks):
        #                 peak_header = [extra + " " +extra_peaks[min_nPeaks+i]["name"] + " centroid",
        #                                extra + " " +extra_peaks[min_nPeaks+i]["name"] + " baseline",
        #                                extra + " " +extra_peaks[min_nPeaks+i]["name"] + " intensity"]
        #                 header.extend(peak_header)
        #
        #             if len(off_mer.keys()) > 0:
        #                 header.extend(self.getOffMerHeaders())
        #
        #             writer.writerow(header)
        #
        #             ### Write data of each group ###
        #             for i in range(len(filelist)):
        #                 if i != 0:
        #                     write_row = ["", str(filelist[i])]
        #                 else:
        #                     write_row = [str(g), str(filelist[i]), g_data["integration area width"]]
        #                     for i in range(min_nPeaks):
        #                         peak_detail = [top_peaks[i]["centroid"], bottom_peaks[i]["centroid"],
        #                                        self.average(top_peaks[i]["centroid"], bottom_peaks[i]["centroid"]),
        #                                        top_peaks[i]["baseline"], bottom_peaks[i]["baseline"],
        #                                        self.average(top_peaks[i]["baseline"], bottom_peaks[i]["baseline"]),
        #                                        top_peaks[i]["intensity"], bottom_peaks[i]["intensity"],
        #                                        self.average(top_peaks[i]["intensity"], bottom_peaks[i]["intensity"])
        #                                        ]
        #                         write_row.extend(peak_detail)
        #
        #                     for i in range(max(nTop, nBottom) - min_nPeaks):
        #                         ind = min_nPeaks + i
        #                         peak_detail = [ extra_peaks[ind]["centroid"],
        #                                         extra_peaks[ind]["baseline"],
        #                                         extra_peaks[ind]["intensity"]]
        #                         write_row.extend(peak_detail)
        #
        #                     if len(off_mer.keys()) > 0:
        #                         for t in ["centroid", "baseline", "intensity"]:
        #                             for p in [0,1]:
        #                                 sum_val = 0.
        #                                 for q in ["top_left", "top_right", "bottom_left", "bottom_right"]:
        #                                     write_row.append(off_mer[q][t][p])
        #                                     sum_val += float(off_mer[q][t][p])
        #                                 write_row.append(sum_val/4.)
        #
        #                 writer.writerow(write_row)
        #             writer.writerow([])
        #             writer.writerow([])
        # else:
        # Files are selected by auto-grouping
        with open(self.filename, "w") as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            grpKey = sorted(self.grp_to_key.items(), key=lambda gk: int(gk[0]))
            g1, k1 = grpKey[0]
            peak_data = self.allpeaks_data[k1]

            ### write header for all group ###
            header = ["group", "filename", "integration area width"]
            top_peaks = sorted([d for d in peak_data if d["side"] == "top"], key=lambda d: self.toInt(d["name"]))
            bottom_peaks = sorted([d for d in peak_data if d["side"] == "bottom"], key=lambda d: self.toInt(d["name"]))
            nTop = len(top_peaks)
            nBottom = len(bottom_peaks)
            min_nPeaks = min(nTop, nBottom)

            for i in range(min_nPeaks):
                peak_header = ["top " + top_peaks[i]["name"] + " centroid",
                               "bottom " + bottom_peaks[i]["name"] + " centroid",
                               "average " + bottom_peaks[i]["name"] + " centroid",
                               "top " + top_peaks[i]["name"] + " baseline",
                               "bottom " + bottom_peaks[i]["name"] + " baseline",
                               "average " + bottom_peaks[i]["name"] + " baseline",
                               "top " + top_peaks[i]["name"] + " intensity",
                               "bottom " + bottom_peaks[i]["name"] + " intensity",
                               "average " + bottom_peaks[i]["name"] + " intensity"]
                header.extend(peak_header)
            header.extend(self.getOffMerHeaders())

            writer.writerow(header)
            self.headers = header

            ### Write data for each group ###
            for (g, k) in grpKey:
                off_mer = self.offmer_data[k]
                peak_data = self.allpeaks_data[k]
                g_data = self.gen_data[k]
                filelist = self.keyToImages[k]
                top_peaks = sorted([d for d in peak_data if d["side"] == "top"], key=lambda d: d["name"])
                bottom_peaks = sorted([d for d in peak_data if d["side"] == "bottom"], key=lambda d: d["name"])

                for i in range(len(filelist)):
                    if i != 0:
                        write_row = ["", str(filelist[i])]
                    else:
                        write_row = [str(g), str(filelist[i]), g_data["integration area width"]]
                        for i in range(min_nPeaks):
                            peak_detail = [top_peaks[i]["centroid"], bottom_peaks[i]["centroid"],
                                           self.average(top_peaks[i]["centroid"], bottom_peaks[i]["centroid"]),
                                           top_peaks[i]["baseline"], bottom_peaks[i]["baseline"],
                                           self.average(top_peaks[i]["baseline"], bottom_peaks[i]["baseline"]),
                                           top_peaks[i]["intensity"], bottom_peaks[i]["intensity"],
                                           self.average(top_peaks[i]["intensity"], bottom_peaks[i]["intensity"])]
                            write_row.extend(peak_detail)

                        if len(off_mer.keys()) > 0:
                            for h in ["centroid", "baseline", "intensity"]:
                                for p in [0,1]:
                                    sum_val = 0.
                                    for q in ["top_left", "top_right", "bottom_left", "bottom_right"]:
                                        write_row.append(off_mer[q][h][p])
                                        sum_val += float(off_mer[q][h][p])
                                    write_row.append(sum_val/4.)

                    writer.writerow(write_row)

    def average(self, s1, s2):
        # Find average between 2 float string. Remove '_' if necessary
        rejected = False
        if "_" in s1:
            rejected = True
            s1 = s1.lstrip("_")
        if "_" in s2:
            rejected = True
            s2 = s2.lstrip("_")

        result = str(1.*(float(s1)+float(s2))/2.)
        if rejected :
            result = "_" + result
        return result


#!/usr/bin/python

"""
Copyright 1999 Illinois Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL ILLINOIS INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of Illinois Institute
of Technology shall not be used in advertising or otherwise to promote
the sale, use or other dealings in this Software without prior written
authorization from Illinois Institute of Technology.
"""
import sys
from musclex import __version__
from musclex.ui.pyqt_utils import *
from musclex.utils.exception_handler import handlers

if sys.platform in handlers:
    sys.excepthook = handlers[sys.platform]

def main(arguments=None):
    if arguments is None:
        arguments = sys.argv

    run = True
    if len(arguments) == 2:
        prog = arguments[1]
        if prog == 'eq':
            app = QApplication(sys.argv)
            from musclex.ui.EQStartWindow import EQStartWindow
            myapp = EQStartWindow()
            sys.exit(app.exec_())
        elif prog == 'qf':
            app = QApplication(sys.argv)
            from musclex.ui.QuadrantFoldingGUI import QuadrantFoldingGUI
            myapp = QuadrantFoldingGUI()
            sys.exit(app.exec_())
        elif prog == 'di':
            app = QApplication(sys.argv)
            from musclex.ui.ScanningDiffractionGUI import ScanningDiffractionGUI
            myapp = ScanningDiffractionGUI()
            sys.exit(app.exec_())
        elif prog == 'dc':
            from musclex.ui.diffraction_centroids import DiffractionCentroidStartWindow
            app = QApplication(sys.argv)
            myapp = DiffractionCentroidStartWindow()
            sys.exit(app.exec_())
        elif prog == 'ddf':
            from musclex.ui.ddf_processor import DDFWindow
            app = QApplication(sys.argv)
            myapp = DDFWindow()
            sys.exit(app.exec_())
        elif prog == 'pt':
            from musclex.ui.ProjectionTracesGUI import ProjectionTracesGUI
            app = QApplication(sys.argv)
            myapp = ProjectionTracesGUI()
            sys.exit(app.exec_())
        elif prog == 'im':
            from musclex.ui.ImageMerger import ImageMergerGUI
            app = QApplication(sys.argv)
            myapp = ImageMergerGUI()
            sys.exit(app.exec_())
        else:
            run = False
    else:
        run = False

    if not run:
        print("\nYou're using Muscle X version "+str(__version__))
        print("\nPlease specify the program shortcut that you want to run")
        print("")
        print("  $ musclex [--program]")
        print("")
        print("          eq - Equator")
        print("          qf - Quadrant Folding")
        print("          pt - Projection Traces")
        print("          di - Scanning Diffraction")
        print("          im - Image Merger")
        print("          dc - Diffraction Centroids")
        print("          ddf - DDF Processor")
        print("")
        print("For example,")
        print("\t$ musclex eq")
        print("")
        print("More details : https://musclex.readthedocs.io")
        print("Submit Feedback or issues : https://www.github.com/biocatiit/musclex/issues\n\n")

if __name__ == "__main__":
    main(sys.argv)
